import json
import os.path
import unittest

import requests_mock

import tableauserverclient as TSC
from tableauserverclient.server.endpoint.exceptions import GraphQLError

TEST_ASSET_DIR = os.path.join(os.path.dirname(__file__), "assets")

METADATA_QUERY_SUCCESS = os.path.join(TEST_ASSET_DIR, "metadata_query_success.json")
METADATA_QUERY_ERROR = os.path.join(TEST_ASSET_DIR, "metadata_query_error.json")
EXPECTED_PAGED_DICT = os.path.join(TEST_ASSET_DIR, "metadata_query_expected_dict.dict")

METADATA_PAGE_1 = os.path.join(TEST_ASSET_DIR, "metadata_paged_1.json")
METADATA_PAGE_2 = os.path.join(TEST_ASSET_DIR, "metadata_paged_2.json")
METADATA_PAGE_3 = os.path.join(TEST_ASSET_DIR, "metadata_paged_3.json")

EXPECTED_DICT = {
    "publishedDatasources": [
        {"id": "01cf92b2-2d17-b656-fc48-5c25ef6d5352", "name": "Batters (TestV1)"},
        {"id": "020ae1cd-c356-f1ad-a846-b0094850d22a", "name": "SharePoint_List_sharepoint2010.test.tsi.lan"},
        {"id": "061493a0-c3b2-6f39-d08c-bc3f842b44af", "name": "Batters_mongodb"},
        {"id": "089fe515-ad2f-89bc-94bd-69f55f69a9c2", "name": "Sample - Superstore"},
    ]
}

EXPECTED_DICT_ERROR = [{"message": "Reached time limit of PT5S for query execution.", "path": None, "extensions": None}]


class MetadataTests(unittest.TestCase):
    def setUp(self):
        self.server = TSC.Server("http://test", False)
        self.baseurl = self.server.metadata.baseurl
        self.server.version = "3.5"

        self.server._site_id = "dad65087-b08b-4603-af4e-2887b8aafc67"
        self.server._auth_token = "j80k54ll2lfMZ0tv97mlPvvSCRyD0DOM"

    def test_metadata_query(self):
        with open(METADATA_QUERY_SUCCESS, "rb") as f:
            response_json = json.loads(f.read().decode())
        with requests_mock.mock() as m:
            m.post(self.baseurl, json=response_json)
            actual = self.server.metadata.query("fake query")

            datasources = actual["data"]

        self.assertDictEqual(EXPECTED_DICT, datasources)

    def test_paged_metadata_query(self):
        with open(EXPECTED_PAGED_DICT, "rb") as f:
            expected = eval(f.read())

        # prepare the 3 pages of results
        with open(METADATA_PAGE_1, "rb") as f:
            result_1 = f.read().decode()
        with open(METADATA_PAGE_2, "rb") as f:
            result_2 = f.read().decode()
        with open(METADATA_PAGE_3, "rb") as f:
            result_3 = f.read().decode()

        with requests_mock.mock() as m:
            m.post(
                self.baseurl,
                [
                    {"text": result_1, "status_code": 200},
                    {"text": result_2, "status_code": 200},
                    {"text": result_3, "status_code": 200},
                ],
            )

            # validation checks for endCursor and hasNextPage,
            # but the query text doesn't matter for the test
            actual = self.server.metadata.paginated_query(
                "fake query endCursor hasNextPage", variables={"first": 1, "afterToken": None}
            )

        self.assertDictEqual(expected, actual)

    def test_metadata_query_ignore_error(self):
        with open(METADATA_QUERY_ERROR, "rb") as f:
            response_json = json.loads(f.read().decode())
        with requests_mock.mock() as m:
            m.post(self.baseurl, json=response_json)
            actual = self.server.metadata.query("fake query")
            datasources = actual["data"]

        self.assertNotEqual(actual.get("errors", None), None)
        self.assertListEqual(EXPECTED_DICT_ERROR, actual["errors"])
        self.assertDictEqual(EXPECTED_DICT, datasources)

    def test_metadata_query_abort_on_error(self):
        with open(METADATA_QUERY_ERROR, "rb") as f:
            response_json = json.loads(f.read().decode())
        with requests_mock.mock() as m:
            m.post(self.baseurl, json=response_json)

            with self.assertRaises(GraphQLError) as e:
                self.server.metadata.query("fake query", abort_on_error=True)
                self.assertListEqual(e.error, EXPECTED_DICT_ERROR)



import xml.etree.ElementTree as ET
from datetime import datetime
from typing import List, Optional, Set

from tableauserverclient.datetime_helpers import parse_datetime
from .property_decorators import property_is_boolean, property_is_datetime
from .tag_item import TagItem


class MetricItem(object):
    def __init__(self, name: Optional[str] = None):
        self._id: Optional[str] = None
        self._name: Optional[str] = name
        self._description: Optional[str] = None
        self._webpage_url: Optional[str] = None
        self._created_at: Optional[datetime] = None
        self._updated_at: Optional[datetime] = None
        self._suspended: Optional[bool] = None
        self._project_id: Optional[str] = None
        self._project_name: Optional[str] = None
        self._owner_id: Optional[str] = None
        self._view_id: Optional[str] = None
        self._initial_tags: Set[str] = set()
        self.tags: Set[str] = set()

    @property
    def id(self) -> Optional[str]:
        return self._id

    @id.setter
    def id(self, value: Optional[str]) -> None:
        self._id = value

    @property
    def name(self) -> Optional[str]:
        return self._name

    @name.setter
    def name(self, value: Optional[str]) -> None:
        self._name = value

    @property
    def description(self) -> Optional[str]:
        return self._description

    @description.setter
    def description(self, value: Optional[str]) -> None:
        self._description = value

    @property
    def webpage_url(self) -> Optional[str]:
        return self._webpage_url

    @property
    def created_at(self) -> Optional[datetime]:
        return self._created_at

    @created_at.setter
    @property_is_datetime
    def created_at(self, value: "datetime") -> None:
        self._created_at = value

    @property
    def updated_at(self) -> Optional[datetime]:
        return self._updated_at

    @updated_at.setter
    @property_is_datetime
    def updated_at(self, value: "datetime") -> None:
        self._updated_at = value

    @property
    def suspended(self) -> Optional[bool]:
        return self._suspended

    @suspended.setter
    @property_is_boolean
    def suspended(self, value: bool) -> None:
        self._suspended = value

    @property
    def project_id(self) -> Optional[str]:
        return self._project_id

    @project_id.setter
    def project_id(self, value: Optional[str]) -> None:
        self._project_id = value

    @property
    def project_name(self) -> Optional[str]:
        return self._project_name

    @project_name.setter
    def project_name(self, value: Optional[str]) -> None:
        self._project_name = value

    @property
    def owner_id(self) -> Optional[str]:
        return self._owner_id

    @owner_id.setter
    def owner_id(self, value: Optional[str]) -> None:
        self._owner_id = value

    @property
    def view_id(self) -> Optional[str]:
        return self._view_id

    @view_id.setter
    def view_id(self, value: Optional[str]) -> None:
        self._view_id = value

    def __repr__(self):
        return "<MetricItem# name={_name} id={_id} owner_id={_owner_id}>".format(**vars(self))

    @classmethod
    def from_response(
        cls,
        resp: bytes,
        ns,
    ) -> List["MetricItem"]:
        all_metric_items = list()
        parsed_response = ET.fromstring(resp)
        all_metric_xml = parsed_response.findall(".//t:metric", namespaces=ns)
        for metric_xml in all_metric_xml:
            metric_item = cls()
            metric_item._id = metric_xml.get("id", None)
            metric_item._name = metric_xml.get("name", None)
            metric_item._description = metric_xml.get("description", None)
            metric_item._webpage_url = metric_xml.get("webpageUrl", None)
            metric_item._created_at = parse_datetime(metric_xml.get("createdAt", None))
            metric_item._updated_at = parse_datetime(metric_xml.get("updatedAt", None))
            metric_item._suspended = string_to_bool(metric_xml.get("suspended", ""))
            for owner in metric_xml.findall(".//t:owner", namespaces=ns):
                metric_item._owner_id = owner.get("id", None)

            for project in metric_xml.findall(".//t:project", namespaces=ns):
                metric_item._project_id = project.get("id", None)
                metric_item._project_name = project.get("name", None)

            for view in metric_xml.findall(".//t:underlyingView", namespaces=ns):
                metric_item._view_id = view.get("id", None)

            tags = set()
            tags_elem = metric_xml.find(".//t:tags", namespaces=ns)
            if tags_elem is not None:
                all_tags = TagItem.from_xml_element(tags_elem, ns)
                tags = all_tags

            metric_item.tags = tags
            metric_item._initial_tags = tags

            all_metric_items.append(metric_item)
        return all_metric_items


# Used to convert string represented boolean to a boolean type
def string_to_bool(s: str) -> bool:
    return s.lower() == "true"



import os
import unittest

import requests_mock

import tableauserverclient as TSC

TEST_ASSET_DIR = os.path.join(os.path.dirname(__file__), "assets")

CREATE_XML = os.path.join(TEST_ASSET_DIR, "subscription_create.xml")
GET_XML = os.path.join(TEST_ASSET_DIR, "subscription_get.xml")
GET_XML_BY_ID = os.path.join(TEST_ASSET_DIR, "subscription_get_by_id.xml")


class SubscriptionTests(unittest.TestCase):
    def setUp(self) -> None:
        self.server = TSC.Server("http://test", False)
        self.server.version = "2.6"

        # Fake Signin
        self.server._site_id = "dad65087-b08b-4603-af4e-2887b8aafc67"
        self.server._auth_token = "j80k54ll2lfMZ0tv97mlPvvSCRyD0DOM"

        self.baseurl = self.server.subscriptions.baseurl

    def test_get_subscriptions(self) -> None:
        with open(GET_XML, "rb") as f:
            response_xml = f.read().decode("utf-8")
        with requests_mock.mock() as m:
            m.get(self.baseurl, text=response_xml)
            all_subscriptions, pagination_item = self.server.subscriptions.get()

        self.assertEqual(2, pagination_item.total_available)
        subscription = all_subscriptions[0]
        self.assertEqual("382e9a6e-0c08-4a95-b6c1-c14df7bac3e4", subscription.id)
        self.assertEqual("NOT FOUND!", subscription.message)
        self.assertTrue(subscription.attach_image)
        self.assertFalse(subscription.attach_pdf)
        self.assertFalse(subscription.suspended)
        self.assertFalse(subscription.send_if_view_empty)
        self.assertIsNone(subscription.page_orientation)
        self.assertIsNone(subscription.page_size_option)
        self.assertEqual("Not Found Alert", subscription.subject)
        self.assertEqual("cdd716ca-5818-470e-8bec-086885dbadee", subscription.target.id)
        self.assertEqual("View", subscription.target.type)
        self.assertEqual("c0d5fc44-ad8c-4957-bec0-b70ed0f8df1e", subscription.user_id)
        self.assertEqual("7617c389-cdca-4940-a66e-69956fcebf3e", subscription.schedule_id)

        subscription = all_subscriptions[1]
        self.assertEqual("23cb7630-afc8-4c8e-b6cd-83ae0322ec66", subscription.id)
        self.assertEqual("overview", subscription.message)
        self.assertFalse(subscription.attach_image)
        self.assertTrue(subscription.attach_pdf)
        self.assertTrue(subscription.suspended)
        self.assertTrue(subscription.send_if_view_empty)
        self.assertEqual("PORTRAIT", subscription.page_orientation)
        self.assertEqual("A5", subscription.page_size_option)
        self.assertEqual("Last 7 Days", subscription.subject)
        self.assertEqual("2e6b4e8f-22dd-4061-8f75-bf33703da7e5", subscription.target.id)
        self.assertEqual("Workbook", subscription.target.type)
        self.assertEqual("c0d5fc44-ad8c-4957-bec0-b70ed0f8df1e", subscription.user_id)
        self.assertEqual("3407cd38-7b39-4983-86a6-67a1506a5e3f", subscription.schedule_id)

    def test_get_subscription_by_id(self) -> None:
        with open(GET_XML_BY_ID, "rb") as f:
            response_xml = f.read().decode("utf-8")
        with requests_mock.mock() as m:
            m.get(self.baseurl + "/382e9a6e-0c08-4a95-b6c1-c14df7bac3e4", text=response_xml)
            subscription = self.server.subscriptions.get_by_id("382e9a6e-0c08-4a95-b6c1-c14df7bac3e4")

        self.assertEqual("382e9a6e-0c08-4a95-b6c1-c14df7bac3e4", subscription.id)
        self.assertEqual("View", subscription.target.type)
        self.assertEqual("cdd716ca-5818-470e-8bec-086885dbadee", subscription.target.id)
        self.assertEqual("c0d5fc44-ad8c-4957-bec0-b70ed0f8df1e", subscription.user_id)
        self.assertEqual("Not Found Alert", subscription.subject)
        self.assertEqual("7617c389-cdca-4940-a66e-69956fcebf3e", subscription.schedule_id)

    def test_create_subscription(self) -> None:
        with open(CREATE_XML, "rb") as f:
            response_xml = f.read().decode("utf-8")
        with requests_mock.mock() as m:
            m.post(self.baseurl, text=response_xml)

            target_item = TSC.Target("960e61f2-1838-40b2-bba2-340c9492f943", "workbook")
            new_subscription = TSC.SubscriptionItem(
                "subject", "4906c453-d5ec-4972-9ff4-789b629bdfa2", "8d30c8de-0a5f-4bee-b266-c621b4f3eed0", target_item
            )
            new_subscription = self.server.subscriptions.create(new_subscription)

        self.assertEqual("78e9318d-2d29-4d67-b60f-3f2f5fd89ecc", new_subscription.id)
        self.assertEqual("sub_name", new_subscription.subject)
        self.assertEqual("960e61f2-1838-40b2-bba2-340c9492f943", new_subscription.target.id)
        self.assertEqual("Workbook", new_subscription.target.type)
        self.assertEqual("4906c453-d5ec-4972-9ff4-789b629bdfa2", new_subscription.schedule_id)
        self.assertEqual("8d30c8de-0a5f-4bee-b266-c621b4f3eed0", new_subscription.user_id)

    def test_delete_subscription(self) -> None:
        with requests_mock.mock() as m:
            m.delete(self.baseurl + "/78e9318d-2d29-4d67-b60f-3f2f5fd89ecc", status_code=204)
            self.server.subscriptions.delete("78e9318d-2d29-4d67-b60f-3f2f5fd89ecc")



import os
import unittest

import requests_mock

import tableauserverclient as TSC
from tableauserverclient import UserItem, GroupItem, PermissionsRule
from tableauserverclient.datetime_helpers import format_datetime

TEST_ASSET_DIR = os.path.join(os.path.dirname(__file__), "assets")

ADD_TAGS_XML = os.path.join(TEST_ASSET_DIR, "view_add_tags.xml")
GET_XML = os.path.join(TEST_ASSET_DIR, "view_get.xml")
GET_XML_ID = os.path.join(TEST_ASSET_DIR, "view_get_id.xml")
GET_XML_USAGE = os.path.join(TEST_ASSET_DIR, "view_get_usage.xml")
POPULATE_PREVIEW_IMAGE = os.path.join(TEST_ASSET_DIR, "Sample View Image.png")
POPULATE_PDF = os.path.join(TEST_ASSET_DIR, "populate_pdf.pdf")
POPULATE_CSV = os.path.join(TEST_ASSET_DIR, "populate_csv.csv")
POPULATE_EXCEL = os.path.join(TEST_ASSET_DIR, "populate_excel.xlsx")
POPULATE_PERMISSIONS_XML = os.path.join(TEST_ASSET_DIR, "view_populate_permissions.xml")
UPDATE_PERMISSIONS = os.path.join(TEST_ASSET_DIR, "view_update_permissions.xml")
UPDATE_XML = os.path.join(TEST_ASSET_DIR, "workbook_update.xml")


class ViewTests(unittest.TestCase):
    def setUp(self):
        self.server = TSC.Server("http://test", False)
        self.server.version = "3.2"

        # Fake sign in
        self.server._site_id = "dad65087-b08b-4603-af4e-2887b8aafc67"
        self.server._auth_token = "j80k54ll2lfMZ0tv97mlPvvSCRyD0DOM"

        self.baseurl = self.server.views.baseurl
        self.siteurl = self.server.views.siteurl

    def test_get(self) -> None:
        with open(GET_XML, "rb") as f:
            response_xml = f.read().decode("utf-8")
        with requests_mock.mock() as m:
            m.get(self.baseurl, text=response_xml)
            all_views, pagination_item = self.server.views.get()

        self.assertEqual(2, pagination_item.total_available)
        self.assertEqual("d79634e1-6063-4ec9-95ff-50acbf609ff5", all_views[0].id)
        self.assertEqual("ENDANGERED SAFARI", all_views[0].name)
        self.assertEqual("SafariSample/sheets/ENDANGEREDSAFARI", all_views[0].content_url)
        self.assertEqual("3cc6cd06-89ce-4fdc-b935-5294135d6d42", all_views[0].workbook_id)
        self.assertEqual("5de011f8-5aa9-4d5b-b991-f462c8dd6bb7", all_views[0].owner_id)
        self.assertEqual("5241e88d-d384-4fd7-9c2f-648b5247efc5", all_views[0].project_id)
        self.assertEqual(set(["tag1", "tag2"]), all_views[0].tags)
        self.assertIsNone(all_views[0].created_at)
        self.assertIsNone(all_views[0].updated_at)
        self.assertIsNone(all_views[0].sheet_type)

        self.assertEqual("fd252f73-593c-4c4e-8584-c032b8022adc", all_views[1].id)
        self.assertEqual("Overview", all_views[1].name)
        self.assertEqual("Superstore/sheets/Overview", all_views[1].content_url)
        self.assertEqual("6d13b0ca-043d-4d42-8c9d-3f3313ea3a00", all_views[1].workbook_id)
        self.assertEqual("5de011f8-5aa9-4d5b-b991-f462c8dd6bb7", all_views[1].owner_id)
        self.assertEqual("5b534f74-3226-11e8-b47a-cb2e00f738a3", all_views[1].project_id)
        self.assertEqual("2002-05-30T09:00:00Z", format_datetime(all_views[1].created_at))
        self.assertEqual("2002-06-05T08:00:59Z", format_datetime(all_views[1].updated_at))
        self.assertEqual("story", all_views[1].sheet_type)

    def test_get_by_id(self) -> None:
        with open(GET_XML_ID, "rb") as f:
            response_xml = f.read().decode("utf-8")
        with requests_mock.mock() as m:
            m.get(self.baseurl + "/d79634e1-6063-4ec9-95ff-50acbf609ff5", text=response_xml)
            view = self.server.views.get_by_id("d79634e1-6063-4ec9-95ff-50acbf609ff5")

        self.assertEqual("d79634e1-6063-4ec9-95ff-50acbf609ff5", view.id)
        self.assertEqual("ENDANGERED SAFARI", view.name)
        self.assertEqual("SafariSample/sheets/ENDANGEREDSAFARI", view.content_url)
        self.assertEqual("3cc6cd06-89ce-4fdc-b935-5294135d6d42", view.workbook_id)
        self.assertEqual("5de011f8-5aa9-4d5b-b991-f462c8dd6bb7", view.owner_id)
        self.assertEqual("5241e88d-d384-4fd7-9c2f-648b5247efc5", view.project_id)
        self.assertEqual(set(["tag1", "tag2"]), view.tags)
        self.assertEqual("2002-05-30T09:00:00Z", format_datetime(view.created_at))
        self.assertEqual("2002-06-05T08:00:59Z", format_datetime(view.updated_at))
        self.assertEqual("story", view.sheet_type)

    def test_get_by_id_missing_id(self) -> None:
        self.assertRaises(TSC.MissingRequiredFieldError, self.server.views.get_by_id, None)

    def test_get_with_usage(self) -> None:
        with open(GET_XML_USAGE, "rb") as f:
            response_xml = f.read().decode("utf-8")
        with requests_mock.mock() as m:
            m.get(self.baseurl + "?includeUsageStatistics=true", text=response_xml)
            all_views, pagination_item = self.server.views.get(usage=True)

        self.assertEqual("d79634e1-6063-4ec9-95ff-50acbf609ff5", all_views[0].id)
        self.assertEqual(7, all_views[0].total_views)
        self.assertIsNone(all_views[0].created_at)
        self.assertIsNone(all_views[0].updated_at)
        self.assertIsNone(all_views[0].sheet_type)

        self.assertEqual("fd252f73-593c-4c4e-8584-c032b8022adc", all_views[1].id)
        self.assertEqual(13, all_views[1].total_views)
        self.assertEqual("2002-05-30T09:00:00Z", format_datetime(all_views[1].created_at))
        self.assertEqual("2002-06-05T08:00:59Z", format_datetime(all_views[1].updated_at))
        self.assertEqual("story", all_views[1].sheet_type)

    def test_get_with_usage_and_filter(self) -> None:
        with open(GET_XML_USAGE, "rb") as f:
            response_xml = f.read().decode("utf-8")
        with requests_mock.mock() as m:
            m.get(self.baseurl + "?includeUsageStatistics=true&filter=name:in:[foo,bar]", text=response_xml)
            options = TSC.RequestOptions()
            options.filter.add(
                TSC.Filter(TSC.RequestOptions.Field.Name, TSC.RequestOptions.Operator.In, ["foo", "bar"])
            )
            all_views, pagination_item = self.server.views.get(req_options=options, usage=True)

        self.assertEqual("ENDANGERED SAFARI", all_views[0].name)
        self.assertEqual(7, all_views[0].total_views)
        self.assertEqual("Overview", all_views[1].name)
        self.assertEqual(13, all_views[1].total_views)

    def test_get_before_signin(self) -> None:
        self.server._auth_token = None
        self.assertRaises(TSC.NotSignedInError, self.server.views.get)

    def test_populate_preview_image(self) -> None:
        with open(POPULATE_PREVIEW_IMAGE, "rb") as f:
            response = f.read()
        with requests_mock.mock() as m:
            m.get(
                self.siteurl + "/workbooks/3cc6cd06-89ce-4fdc-b935-5294135d6d42/"
                "views/d79634e1-6063-4ec9-95ff-50acbf609ff5/previewImage",
                content=response,
            )
            single_view = TSC.ViewItem()
            single_view._id = "d79634e1-6063-4ec9-95ff-50acbf609ff5"
            single_view._workbook_id = "3cc6cd06-89ce-4fdc-b935-5294135d6d42"
            self.server.views.populate_preview_image(single_view)
            self.assertEqual(response, single_view.preview_image)

    def test_populate_preview_image_missing_id(self) -> None:
        single_view = TSC.ViewItem()
        single_view._id = "d79634e1-6063-4ec9-95ff-50acbf609ff5"
        self.assertRaises(TSC.MissingRequiredFieldError, self.server.views.populate_preview_image, single_view)

        single_view._id = None
        single_view._workbook_id = "3cc6cd06-89ce-4fdc-b935-5294135d6d42"
        self.assertRaises(TSC.MissingRequiredFieldError, self.server.views.populate_preview_image, single_view)

    def test_populate_image(self) -> None:
        with open(POPULATE_PREVIEW_IMAGE, "rb") as f:
            response = f.read()
        with requests_mock.mock() as m:
            m.get(self.baseurl + "/d79634e1-6063-4ec9-95ff-50acbf609ff5/image", content=response)
            single_view = TSC.ViewItem()
            single_view._id = "d79634e1-6063-4ec9-95ff-50acbf609ff5"
            self.server.views.populate_image(single_view)
            self.assertEqual(response, single_view.image)

    def test_populate_image_with_options(self) -> None:
        with open(POPULATE_PREVIEW_IMAGE, "rb") as f:
            response = f.read()
        with requests_mock.mock() as m:
            m.get(
                self.baseurl + "/d79634e1-6063-4ec9-95ff-50acbf609ff5/image?resolution=high&maxAge=10", content=response
            )
            single_view = TSC.ViewItem()
            single_view._id = "d79634e1-6063-4ec9-95ff-50acbf609ff5"
            req_option = TSC.ImageRequestOptions(imageresolution=TSC.ImageRequestOptions.Resolution.High, maxage=10)
            self.server.views.populate_image(single_view, req_option)
            self.assertEqual(response, single_view.image)

    def test_populate_pdf(self) -> None:
        with open(POPULATE_PDF, "rb") as f:
            response = f.read()
        with requests_mock.mock() as m:
            m.get(
                self.baseurl + "/d79634e1-6063-4ec9-95ff-50acbf609ff5/pdf?type=letter&orientation=portrait&maxAge=5",
                content=response,
            )
            single_view = TSC.ViewItem()
            single_view._id = "d79634e1-6063-4ec9-95ff-50acbf609ff5"

            size = TSC.PDFRequestOptions.PageType.Letter
            orientation = TSC.PDFRequestOptions.Orientation.Portrait
            req_option = TSC.PDFRequestOptions(size, orientation, 5)

            self.server.views.populate_pdf(single_view, req_option)
            self.assertEqual(response, single_view.pdf)

    def test_populate_csv(self) -> None:
        with open(POPULATE_CSV, "rb") as f:
            response = f.read()
        with requests_mock.mock() as m:
            m.get(self.baseurl + "/d79634e1-6063-4ec9-95ff-50acbf609ff5/data?maxAge=1", content=response)
            single_view = TSC.ViewItem()
            single_view._id = "d79634e1-6063-4ec9-95ff-50acbf609ff5"
            request_option = TSC.CSVRequestOptions(maxage=1)
            self.server.views.populate_csv(single_view, request_option)

            csv_file = b"".join(single_view.csv)
            self.assertEqual(response, csv_file)

    def test_populate_csv_default_maxage(self) -> None:
        with open(POPULATE_CSV, "rb") as f:
            response = f.read()
        with requests_mock.mock() as m:
            m.get(self.baseurl + "/d79634e1-6063-4ec9-95ff-50acbf609ff5/data", content=response)
            single_view = TSC.ViewItem()
            single_view._id = "d79634e1-6063-4ec9-95ff-50acbf609ff5"
            self.server.views.populate_csv(single_view)

            csv_file = b"".join(single_view.csv)
            self.assertEqual(response, csv_file)

    def test_populate_image_missing_id(self) -> None:
        single_view = TSC.ViewItem()
        single_view._id = None
        self.assertRaises(TSC.MissingRequiredFieldError, self.server.views.populate_image, single_view)

    def test_populate_permissions(self) -> None:
        with open(POPULATE_PERMISSIONS_XML, "rb") as f:
            response_xml = f.read().decode("utf-8")
        with requests_mock.mock() as m:
            m.get(self.baseurl + "/e490bec4-2652-4fda-8c4e-f087db6fa328/permissions", text=response_xml)
            single_view = TSC.ViewItem()
            single_view._id = "e490bec4-2652-4fda-8c4e-f087db6fa328"

            self.server.views.populate_permissions(single_view)
            permissions = single_view.permissions

            self.assertEqual(permissions[0].grantee.tag_name, "group")
            self.assertEqual(permissions[0].grantee.id, "c8f2773a-c83a-11e8-8c8f-33e6d787b506")
            self.assertDictEqual(
                permissions[0].capabilities,
                {
                    TSC.Permission.Capability.ViewComments: TSC.Permission.Mode.Allow,
                    TSC.Permission.Capability.Read: TSC.Permission.Mode.Allow,
                    TSC.Permission.Capability.AddComment: TSC.Permission.Mode.Allow,
                    TSC.Permission.Capability.ExportData: TSC.Permission.Mode.Allow,
                    TSC.Permission.Capability.ExportImage: TSC.Permission.Mode.Allow,
                },
            )

    def test_add_permissions(self) -> None:
        with open(UPDATE_PERMISSIONS, "rb") as f:
            response_xml = f.read().decode("utf-8")

        single_view = TSC.ViewItem()
        single_view._id = "21778de4-b7b9-44bc-a599-1506a2639ace"

        bob = UserItem.as_reference("7c37ee24-c4b1-42b6-a154-eaeab7ee330a")
        group_of_people = GroupItem.as_reference("5e5e1978-71fa-11e4-87dd-7382f5c437af")

        new_permissions = [PermissionsRule(bob, {"Write": "Allow"}), PermissionsRule(group_of_people, {"Read": "Deny"})]

        with requests_mock.mock() as m:
            m.put(self.baseurl + "/21778de4-b7b9-44bc-a599-1506a2639ace/permissions", text=response_xml)
            permissions = self.server.views.update_permissions(single_view, new_permissions)

        self.assertEqual(permissions[0].grantee.tag_name, "group")
        self.assertEqual(permissions[0].grantee.id, "5e5e1978-71fa-11e4-87dd-7382f5c437af")
        self.assertDictEqual(permissions[0].capabilities, {TSC.Permission.Capability.Read: TSC.Permission.Mode.Deny})

        self.assertEqual(permissions[1].grantee.tag_name, "user")
        self.assertEqual(permissions[1].grantee.id, "7c37ee24-c4b1-42b6-a154-eaeab7ee330a")
        self.assertDictEqual(permissions[1].capabilities, {TSC.Permission.Capability.Write: TSC.Permission.Mode.Allow})

    def test_update_tags(self) -> None:
        with open(ADD_TAGS_XML, "rb") as f:
            add_tags_xml = f.read().decode("utf-8")
        with open(UPDATE_XML, "rb") as f:
            update_xml = f.read().decode("utf-8")
        with requests_mock.mock() as m:
            m.put(self.baseurl + "/d79634e1-6063-4ec9-95ff-50acbf609ff5/tags", text=add_tags_xml)
            m.delete(self.baseurl + "/d79634e1-6063-4ec9-95ff-50acbf609ff5/tags/b", status_code=204)
            m.delete(self.baseurl + "/d79634e1-6063-4ec9-95ff-50acbf609ff5/tags/d", status_code=204)
            m.put(self.baseurl + "/d79634e1-6063-4ec9-95ff-50acbf609ff5", text=update_xml)
            single_view = TSC.ViewItem()
            single_view._id = "d79634e1-6063-4ec9-95ff-50acbf609ff5"
            single_view._initial_tags.update(["a", "b", "c", "d"])
            single_view.tags.update(["a", "c", "e"])
            updated_view = self.server.views.update(single_view)

        self.assertEqual(single_view.tags, updated_view.tags)
        self.assertEqual(single_view._initial_tags, updated_view._initial_tags)

    def test_populate_excel(self) -> None:
        self.server.version = "3.8"
        self.baseurl = self.server.views.baseurl
        with open(POPULATE_EXCEL, "rb") as f:
            response = f.read()
        with requests_mock.mock() as m:
            m.get(self.baseurl + "/d79634e1-6063-4ec9-95ff-50acbf609ff5/crosstab/excel?maxAge=1", content=response)
            single_view = TSC.ViewItem()
            single_view._id = "d79634e1-6063-4ec9-95ff-50acbf609ff5"
            request_option = TSC.ExcelRequestOptions(maxage=1)
            self.server.views.populate_excel(single_view, request_option)

            excel_file = b"".join(single_view.excel)
            self.assertEqual(response, excel_file)



import unittest

import tableauserverclient as TSC


class ProjectModelTests(unittest.TestCase):
    def test_invalid_name(self):
        self.assertRaises(ValueError, TSC.ProjectItem, None)
        self.assertRaises(ValueError, TSC.ProjectItem, "")
        project = TSC.ProjectItem("proj")
        with self.assertRaises(ValueError):
            project.name = None

        with self.assertRaises(ValueError):
            project.name = ""

    def test_invalid_content_permissions(self):
        project = TSC.ProjectItem("proj")
        with self.assertRaises(ValueError):
            project.content_permissions = "Hello"

    def test_parent_id(self):
        project = TSC.ProjectItem("proj")
        project.parent_id = "foo"
        self.assertEqual(project.parent_id, "foo")

    def test_owner_id(self):
        project = TSC.ProjectItem("proj")
        with self.assertRaises(NotImplementedError):
            project.owner_id = "new_owner"



####
# This script demonstrates how to create schedules using the Tableau
# Server Client.
#
# To run the script, you must have installed Python 3.7 or later.
####


import argparse
import logging

from datetime import time

import tableauserverclient as TSC


def main():
    parser = argparse.ArgumentParser(description="Creates sample schedules for each type of frequency.")
    # Common options; please keep those in sync across all samples
    parser.add_argument("--server", "-s", required=True, help="server address")
    parser.add_argument("--site", "-S", help="site name")
    parser.add_argument(
        "--token-name", "-p", required=True, help="name of the personal access token used to sign into the server"
    )
    parser.add_argument(
        "--token-value", "-v", required=True, help="value of the personal access token used to sign into the server"
    )
    parser.add_argument(
        "--logging-level",
        "-l",
        choices=["debug", "info", "error"],
        default="error",
        help="desired logging level (set to error by default)",
    )
    # Options specific to this sample:
    # This sample has no additional options, yet. If you add some, please add them here

    args = parser.parse_args()

    # Set logging level based on user input, or error by default
    logging_level = getattr(logging, args.logging_level.upper())
    logging.basicConfig(level=logging_level)

    tableau_auth = TSC.PersonalAccessTokenAuth(args.token_name, args.token_value, site_id=args.site)
    server = TSC.Server(args.server, use_server_version=False)
    server.add_http_options({"verify": False})
    server.use_server_version()
    with server.auth.sign_in(tableau_auth):
        # Hourly Schedule
        # This schedule will run every 2 hours between 2:30AM and 11:00PM
        hourly_interval = TSC.HourlyInterval(start_time=time(2, 30), end_time=time(23, 0), interval_value=2)

        hourly_schedule = TSC.ScheduleItem(
            "Hourly-Schedule",
            50,
            TSC.ScheduleItem.Type.Extract,
            TSC.ScheduleItem.ExecutionOrder.Parallel,
            hourly_interval,
        )
        try:
            hourly_schedule = server.schedules.create(hourly_schedule)
            print("Hourly schedule created (ID: {}).".format(hourly_schedule.id))
        except Exception as e:
            print(e)

        # Daily Schedule
        # This schedule will run every day at 5AM
        daily_interval = TSC.DailyInterval(start_time=time(5))
        daily_schedule = TSC.ScheduleItem(
            "Daily-Schedule",
            60,
            TSC.ScheduleItem.Type.Subscription,
            TSC.ScheduleItem.ExecutionOrder.Serial,
            daily_interval,
        )
        try:
            daily_schedule = server.schedules.create(daily_schedule)
            print("Daily schedule created (ID: {}).".format(daily_schedule.id))
        except Exception as e:
            print(e)

        # Weekly Schedule
        # This schedule will wun every Monday, Wednesday, and Friday at 7:15PM
        weekly_interval = TSC.WeeklyInterval(
            time(19, 15), TSC.IntervalItem.Day.Monday, TSC.IntervalItem.Day.Wednesday, TSC.IntervalItem.Day.Friday
        )
        weekly_schedule = TSC.ScheduleItem(
            "Weekly-Schedule",
            70,
            TSC.ScheduleItem.Type.Extract,
            TSC.ScheduleItem.ExecutionOrder.Serial,
            weekly_interval,
        )
        try:
            weekly_schedule = server.schedules.create(weekly_schedule)
            print("Weekly schedule created (ID: {}).".format(weekly_schedule.id))
        except Exception as e:
            print(e)
            options = TSC.RequestOptions()
            options.filter.add(
                TSC.Filter(TSC.RequestOptions.Field.Name, TSC.RequestOptions.Operator.Equals, "Weekly Schedule")
            )
            schedules, _ = server.schedules.get(req_options=options)
            weekly_schedule = schedules[0]
            print(weekly_schedule)

        # Monthly Schedule
        # This schedule will run on the 15th of every month at 11:30PM
        monthly_interval = TSC.MonthlyInterval(start_time=time(23, 30), interval_value=15)
        monthly_schedule = TSC.ScheduleItem(
            "Monthly-Schedule",
            80,
            TSC.ScheduleItem.Type.Subscription,
            TSC.ScheduleItem.ExecutionOrder.Parallel,
            monthly_interval,
        )
        try:
            monthly_schedule = server.schedules.create(monthly_schedule)
            print("Monthly schedule created (ID: {}).".format(monthly_schedule.id))
        except Exception as e:
            print(e)

        # Now fetch the weekly schedule by id
        fetched_schedule = server.schedules.get_by_id(weekly_schedule.id)
        fetched_interval = fetched_schedule.interval_item
        print("Fetched back our weekly schedule, it shows interval ", fetched_interval)


if __name__ == "__main__":
    main()



import logging

import requests
import urllib3

from defusedxml.ElementTree import fromstring, ParseError
from packaging.version import Version

from . import CustomViews
from .endpoint import (
    Sites,
    Views,
    Users,
    Groups,
    Workbooks,
    Datasources,
    Projects,
    Auth,
    Schedules,
    ServerInfo,
    Tasks,
    Subscriptions,
    Jobs,
    Metadata,
    Databases,
    Tables,
    Flows,
    Webhooks,
    DataAccelerationReport,
    Favorites,
    DataAlerts,
    Fileuploads,
    FlowRuns,
    Metrics,
    Endpoint,
)
from .endpoint.exceptions import (
    ServerInfoEndpointNotFoundError,
    EndpointUnavailableError,
)
from .exceptions import NotSignedInError
from ..namespace import Namespace


_PRODUCT_TO_REST_VERSION = {
    "10.0": "2.3",
    "9.3": "2.2",
    "9.2": "2.1",
    "9.1": "2.0",
    "9.0": "2.0",
}

minimum_supported_server_version = "2.3"
default_server_version = "2.4"  # first version that dropped the legacy auth endpoint


class Server(object):
    class PublishMode:
        Append = "Append"
        Overwrite = "Overwrite"
        CreateNew = "CreateNew"

    def __init__(self, server_address, use_server_version=False, http_options=None, session_factory=None):
        self._auth_token = None
        self._site_id = None
        self._user_id = None

        # TODO: this needs to change to default to https, but without breaking existing code
        if not server_address.startswith("http://") and not server_address.startswith("https://"):
            server_address = "http://" + server_address

        self._server_address: str = server_address
        self._session_factory = session_factory or requests.session

        self.auth = Auth(self)
        self.views = Views(self)
        self.users = Users(self)
        self.sites = Sites(self)
        self.groups = Groups(self)
        self.jobs = Jobs(self)
        self.workbooks = Workbooks(self)
        self.datasources = Datasources(self)
        self.favorites = Favorites(self)
        self.flows = Flows(self)
        self.projects = Projects(self)
        self.schedules = Schedules(self)
        self.server_info = ServerInfo(self)
        self.tasks = Tasks(self)
        self.subscriptions = Subscriptions(self)
        self.metadata = Metadata(self)
        self.databases = Databases(self)
        self.tables = Tables(self)
        self.webhooks = Webhooks(self)
        self.data_acceleration_report = DataAccelerationReport(self)
        self.data_alerts = DataAlerts(self)
        self.fileuploads = Fileuploads(self)
        self._namespace = Namespace()
        self.flow_runs = FlowRuns(self)
        self.metrics = Metrics(self)
        self.custom_views = CustomViews(self)

        self.logger = logging.getLogger("TSC.server")

        self._session = self._session_factory()
        self._http_options = dict()  # must set this before making a server call
        if http_options:
            self.add_http_options(http_options)

        self.validate_connection_settings()  # does not make an actual outgoing request

        self.version = default_server_version
        if use_server_version:
            self.use_server_version()  # this makes a server call

    def validate_connection_settings(self):
        try:
            Endpoint(self).set_parameters(self._http_options, None, None, None, None)
            if not self._server_address.startswith("http://") and not self._server_address.startswith("https://"):
                self._server_address = "http://" + self._server_address
            self._session.prepare_request(requests.Request("GET", url=self._server_address, params=self._http_options))
        except Exception as req_ex:
            raise ValueError("Server connection settings not valid", req_ex)

    def __repr__(self):
        return "<TableauServerClient [Connection: {}, {}]>".format(self.baseurl, self.server_info.serverInfo)

    def add_http_options(self, options_dict: dict):
        try:
            self._http_options.update(options_dict)
            if "verify" in options_dict.keys() and self._http_options.get("verify") is False:
                urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)
                # would be nice if you could turn them back on
        except Exception as be:
            # expected errors on invalid input:
            # 'set' object has no attribute 'keys', 'list' object has no attribute 'keys'
            # TypeError: cannot convert dictionary update sequence element #0 to a sequence (input is a tuple)
            raise ValueError("Invalid http options given: {}".format(options_dict))

    def clear_http_options(self):
        self._http_options = dict()

    def _clear_auth(self):
        self._site_id = None
        self._user_id = None
        self._auth_token = None
        self._session = self._session_factory()

    def _set_auth(self, site_id, user_id, auth_token):
        self._site_id = site_id
        self._user_id = user_id
        self._auth_token = auth_token

    def _get_legacy_version(self):
        # the serverInfo call was introduced in 2.4, earlier than that we have this different call
        response = self._session.get(self.server_address + "/auth?format=xml")
        try:
            info_xml = fromstring(response.content)
        except ParseError as parseError:
            self.logger.info(parseError)
            self.logger.info("Could not read server version info. The server may not be running or configured.")
            return self.version
        prod_version = info_xml.find(".//product_version").text
        version = _PRODUCT_TO_REST_VERSION.get(prod_version, minimum_supported_server_version)
        return version

    def _determine_highest_version(self):
        try:
            old_version = self.version
            version = self.server_info.get().rest_api_version
        except ServerInfoEndpointNotFoundError as e:
            self.logger.info("Could not get version info from server: {}{}".format(e.__class__, e))
            version = self._get_legacy_version()
        except EndpointUnavailableError as e:
            self.logger.info("Could not get version info from server: {}{}".format(e.__class__, e))
            version = self._get_legacy_version()
        except Exception as e:
            self.logger.info("Could not get version info from server: {}{}".format(e.__class__, e))
            version = None
        self.logger.info("versions: {}, {}".format(version, old_version))
        return version or old_version

    def use_server_version(self):
        self.version = self._determine_highest_version()

    def use_highest_version(self):
        self.use_server_version()
        self.logger.info("use use_server_version instead", DeprecationWarning)

    def check_at_least_version(self, target: str):
        server_version = Version(self.version or "2.4")
        target_version = Version(target)
        return server_version >= target_version

    def assert_at_least_version(self, comparison: str, reason: str):
        if not self.check_at_least_version(comparison):
            error = "{} is not available in API version {}. Requires {}".format(reason, self.version, comparison)
            raise EndpointUnavailableError(error)

    @property
    def baseurl(self):
        return "{0}/api/{1}".format(self._server_address, str(self.version))

    @property
    def namespace(self):
        return self._namespace()

    @property
    def auth_token(self):
        if self._auth_token is None:
            error = "Missing authentication token. You must sign in first."
            raise NotSignedInError(error)
        return self._auth_token

    @property
    def site_id(self):
        if self._site_id is None:
            error = "Missing site ID. You must sign in first."
            raise NotSignedInError(error)
        return self._site_id

    @property
    def user_id(self):
        if self._user_id is None:
            error = "Missing user ID. You must sign in first."
            raise NotSignedInError(error)
        return self._user_id

    @property
    def server_address(self):
        return self._server_address

    @property
    def http_options(self):
        return self._http_options

    @property
    def session(self):
        return self._session

    def is_signed_in(self):
        return self._auth_token is not None



import requests
import logging
from packaging.version import Version
from functools import wraps
from xml.etree.ElementTree import ParseError
from typing import Any, Callable, Dict, Optional, TYPE_CHECKING

from .exceptions import (
    ServerResponseError,
    InternalServerError,
    NonXMLResponseError,
    EndpointUnavailableError,
)
from tableauserverclient.server.query import QuerySet
from tableauserverclient import helpers, get_versions

if TYPE_CHECKING:
    from ..server import Server
    from requests import Response

logger = logging.getLogger("tableau.endpoint")

Success_codes = [200, 201, 202, 204]

XML_CONTENT_TYPE = "text/xml"
JSON_CONTENT_TYPE = "application/json"

CONTENT_TYPE_HEADER = "content-type"
TABLEAU_AUTH_HEADER = "x-tableau-auth"
USER_AGENT_HEADER = "User-Agent"


class Endpoint(object):
    def __init__(self, parent_srv: "Server"):
        self.parent_srv = parent_srv

    @staticmethod
    def set_parameters(http_options, auth_token, content, content_type, parameters) -> Dict[str, Any]:
        parameters = parameters or {}
        parameters.update(http_options)
        if "headers" not in parameters:
            parameters["headers"] = {}

        if auth_token is not None:
            parameters["headers"][TABLEAU_AUTH_HEADER] = auth_token
        if content_type is not None:
            parameters["headers"][CONTENT_TYPE_HEADER] = content_type

        Endpoint.set_user_agent(parameters)
        if content is not None:
            parameters["data"] = content
        return parameters or {}

    @staticmethod
    def set_user_agent(parameters):
        if USER_AGENT_HEADER not in parameters["headers"]:
            if USER_AGENT_HEADER in parameters:
                parameters["headers"][USER_AGENT_HEADER] = parameters[USER_AGENT_HEADER]
            else:
                # only set the TSC user agent if not already populated
                _client_version: Optional[str] = get_versions()["version"]
                parameters["headers"][USER_AGENT_HEADER] = "Tableau Server Client/{}".format(_client_version)

        # result: parameters["headers"]["User-Agent"] is set
        # return explicitly for testing only
        return parameters

    def _make_request(
        self,
        method: Callable[..., "Response"],
        url: str,
        content: Optional[bytes] = None,
        auth_token: Optional[str] = None,
        content_type: Optional[str] = None,
        parameters: Optional[Dict[str, Any]] = None,
    ) -> "Response":
        parameters = Endpoint.set_parameters(
            self.parent_srv.http_options, auth_token, content, content_type, parameters
        )

        logger.debug("request method {}, url: {}".format(method.__name__, url))
        if content:
            redacted = helpers.strings.redact_xml(content[:1000])
            # logger.debug("request content: {}".format(redacted))

        server_response = method(url, **parameters)
        self._check_status(server_response, url)

        loggable_response = self.log_response_safely(server_response)
        # logger.debug("Server response from {0}:\n\t{1}".format(url, loggable_response))

        if content_type == "application/xml":
            self.parent_srv._namespace.detect(server_response.content)

        return server_response

    def _check_status(self, server_response, url: Optional[str] = None):
        if server_response.status_code >= 500:
            raise InternalServerError(server_response, url)
        elif server_response.status_code not in Success_codes:
            try:
                raise ServerResponseError.from_response(server_response.content, self.parent_srv.namespace, url)
            except ParseError:
                # This will happen if we get a non-success HTTP code that doesn't return an xml error object
                # e.g metadata endpoints, 503 pages, totally different servers
                # we convert this to a better exception and pass through the raw response body
                raise NonXMLResponseError(server_response.content)
            except Exception:
                # anything else re-raise here
                raise

    def log_response_safely(self, server_response: requests.Response) -> str:
        # Checking the content type header prevents eager evaluation of streaming requests.
        content_type = server_response.headers.get("Content-Type")

        # Response.content is a property. Calling it will load the entire response into memory. Checking if the
        # content-type is an octet-stream accomplishes the same goal without eagerly loading content.
        # This check is to determine if the response is a text response (xml or otherwise)
        # so that we do not attempt to log bytes and other binary data.
        loggable_response = "Content type {}".format(content_type)
        if content_type == "application/octet-stream":
            loggable_response = "A stream of type {} [Truncated File Contents]".format(content_type)
        elif server_response.encoding and len(server_response.content) > 0:
            loggable_response = helpers.strings.redact_xml(server_response.content.decode(server_response.encoding))
        return loggable_response

    def get_unauthenticated_request(self, url):
        return self._make_request(self.parent_srv.session.get, url)

    def get_request(self, url, request_object=None, parameters=None):
        if request_object is not None:
            try:
                # Query param delimiters don't need to be encoded for versions before 3.7 (2020.1)
                self.parent_srv.assert_at_least_version("3.7", "Query param encoding")
                parameters = parameters or {}
                parameters["params"] = request_object.get_query_params()
            except EndpointUnavailableError:
                url = request_object.apply_query_params(url)

        return self._make_request(
            self.parent_srv.session.get,
            url,
            auth_token=self.parent_srv.auth_token,
            parameters=parameters,
        )

    def delete_request(self, url):
        # We don't return anything for a delete request
        self._make_request(self.parent_srv.session.delete, url, auth_token=self.parent_srv.auth_token)

    def put_request(self, url, xml_request=None, content_type=XML_CONTENT_TYPE, parameters=None):
        return self._make_request(
            self.parent_srv.session.put,
            url,
            content=xml_request,
            auth_token=self.parent_srv.auth_token,
            content_type=content_type,
            parameters=parameters,
        )

    def post_request(self, url, xml_request, content_type=XML_CONTENT_TYPE, parameters=None):
        return self._make_request(
            self.parent_srv.session.post,
            url,
            content=xml_request,
            auth_token=self.parent_srv.auth_token,
            content_type=content_type,
            parameters=parameters,
        )

    def patch_request(self, url, xml_request, content_type=XML_CONTENT_TYPE, parameters=None):
        return self._make_request(
            self.parent_srv.session.patch,
            url,
            content=xml_request,
            auth_token=self.parent_srv.auth_token,
            content_type=content_type,
            parameters=parameters,
        )


def api(version):
    """Annotate the minimum supported version for an endpoint.

    Checks the version on the server object and compares normalized versions.
    It will raise an exception if the server version is > the version specified.

    Args:
        `version` minimum version that supports the endpoint. String.
    Raises:
        EndpointUnavailableError
    Returns:
        None

    Example:
    >>> @api(version="2.3")
    >>> def get(self, req_options=None):
    >>>     ...
    """

    def _decorator(func):
        @wraps(func)
        def wrapper(self, *args, **kwargs):
            self.parent_srv.assert_at_least_version(version, self.__class__.__name__)
            return func(self, *args, **kwargs)

        return wrapper

    return _decorator


def parameter_added_in(**params):
    """Annotate minimum versions for new parameters or request options on an endpoint.

    The api decorator documents when an endpoint was added, this decorator annotates
    keyword arguments on endpoints that may control functionality added after an endpoint was introduced.

    The REST API will ignore invalid parameters in most cases, so this raises a warning instead of throwing
    an exception.

    Args:
        Key/value pairs of the form `parameter`=`version`. Kwargs.
    Raises:
        UserWarning
    Returns:
        None

    Example:
    >>> @api(version="2.0")
    >>> @parameter_added_in(no_extract='2.5')
    >>> def download(self, workbook_id, filepath=None, extract_only=False):
    >>>     ...
    """

    def _decorator(func):
        @wraps(func)
        def wrapper(self, *args, **kwargs):
            import warnings

            server_ver = Version(self.parent_srv.version or "0.0")
            params_to_check = set(params) & set(kwargs)
            for p in params_to_check:
                min_ver = Version(str(params[p]))
                if server_ver < min_ver:
                    error = "{!r} not available in {}, it will be ignored. Added in {}".format(p, server_ver, min_ver)
                    warnings.warn(error)
            return func(self, *args, **kwargs)

        return wrapper

    return _decorator


class QuerysetEndpoint(Endpoint):
    @api(version="2.0")
    def all(self, *args, **kwargs):
        queryset = QuerySet(self)
        return queryset

    @api(version="2.0")
    def filter(self, *_, **kwargs) -> QuerySet:
        if _:
            raise RuntimeError("Only keyword arguments accepted.")
        queryset = QuerySet(self).filter(**kwargs)
        return queryset

    @api(version="2.0")
    def order_by(self, *args, **kwargs):
        queryset = QuerySet(self).order_by(*args)
        return queryset

    @api(version="2.0")
    def paginate(self, **kwargs):
        queryset = QuerySet(self).paginate(**kwargs)
        return queryset



import logging

from defusedxml.ElementTree import fromstring

from .exceptions import UnpopulatedPropertyError
from .property_decorators import (
    property_is_enum,
    property_not_empty,
    property_is_boolean,
)


class DatabaseItem(object):
    class ContentPermissions:
        LockedToProject = "LockedToDatabase"
        ManagedByOwner = "ManagedByOwner"

    def __init__(self, name, description=None, content_permissions=None):
        self._id = None
        self.name = name
        self.description = description
        self.content_permissions = content_permissions
        self._certified = None
        self._certification_note = None
        self._contact_id = None

        self._connector_url = None
        self._connection_type = None
        self._embedded = None
        self._file_extension = None
        self._file_id = None
        self._file_path = None
        self._host_name = None
        self._metadata_type = None
        self._mime_type = None
        self._port = None
        self._provider = None
        self._request_url = None

        self._permissions = None
        self._default_table_permissions = None

        self._data_quality_warnings = None

        self._tables = None  # Not implemented yet

    @property
    def dqws(self):
        if self._data_quality_warnings is None:
            error = "Project item must be populated with permissions first."
            raise UnpopulatedPropertyError(error)
        return self._data_quality_warnings()

    @property
    def content_permissions(self):
        return self._content_permissions

    @content_permissions.setter
    @property_is_enum(ContentPermissions)
    def content_permissions(self, value):
        self._content_permissions = value

    @property
    def permissions(self):
        if self._permissions is None:
            error = "Project item must be populated with permissions first."
            raise UnpopulatedPropertyError(error)
        return self._permissions()

    @property
    def default_table_permissions(self):
        if self._default_table_permissions is None:
            error = "Project item must be populated with permissions first."
            raise UnpopulatedPropertyError(error)
        return self._default_table_permissions()

    @property
    def id(self):
        return self._id

    @property
    def name(self):
        return self._name

    @name.setter
    @property_not_empty
    def name(self, value):
        self._name = value

    @property
    def description(self):
        return self._description

    @description.setter
    def description(self, value):
        self._description = value

    @property
    def embedded(self):
        return self._embedded

    @property
    def certified(self):
        return self._certified

    @certified.setter
    @property_is_boolean
    def certified(self, value):
        self._certified = value

    @property
    def certification_note(self):
        return self._certification_note

    @certification_note.setter
    def certification_note(self, value):
        self._certification_note = value

    @property
    def metadata_type(self):
        return self._metadata_type

    @property
    def host_name(self):
        return self._host_name

    @property
    def port(self):
        return self._port

    @property
    def file_path(self):
        return self._file_path

    @property
    def provider(self):
        return self._provider

    @property
    def mime_type(self):
        return self._mime_type

    @property
    def connector_url(self):
        return self._connector_url

    @property
    def connection_type(self):
        return self._connection_type

    @property
    def request_url(self):
        return self._request_url

    @property
    def file_extension(self):
        return self._file_extension

    @property
    def file_id(self):
        return self._file_id

    @property
    def contact_id(self):
        return self._contact_id

    @contact_id.setter
    def contact_id(self, value):
        self._contact_id = value

    @property
    def tables(self):
        if self._tables is None:
            error = "Database must be populated with tables first."
            raise UnpopulatedPropertyError(error)
        #  Each call to `.tables` should create a new pager, this just runs the callable
        return self._tables()

    def _set_values(self, database_values):
        # ID & Settable
        if "id" in database_values:
            self._id = database_values["id"]

        if "contact" in database_values:
            self._contact_id = database_values["contact"]["id"]

        if "name" in database_values:
            self._name = database_values["name"]

        if "description" in database_values:
            self._description = database_values["description"]

        if "isCertified" in database_values:
            self._certified = string_to_bool(database_values["isCertified"])

        if "certificationNote" in database_values:
            self._certification_note = database_values["certificationNote"]

        # Not settable, alphabetical

        if "connectionType" in database_values:
            self._connection_type = database_values["connectionType"]

        if "connectorUrl" in database_values:
            self._connector_url = database_values["connectorUrl"]

        if "contentPermissions" in database_values:
            self._content_permissions = database_values["contentPermissions"]

        if "isEmbedded" in database_values:
            self._embedded = string_to_bool(database_values["isEmbedded"])

        if "fileExtension" in database_values:
            self._file_extension = database_values["fileExtension"]

        if "fileId" in database_values:
            self._file_id = database_values["fileId"]

        if "filePath" in database_values:
            self._file_path = database_values["filePath"]

        if "hostName" in database_values:
            self._host_name = database_values["hostName"]

        if "mimeType" in database_values:
            self._mime_type = database_values["mimeType"]

        if "port" in database_values:
            self._port = int(database_values["port"])

        if "provider" in database_values:
            self._provider = database_values["provider"]

        if "requestUrl" in database_values:
            self._request_url = database_values["requestUrl"]

        if "type" in database_values:
            self._metadata_type = database_values["type"]

    def _set_permissions(self, permissions):
        self._permissions = permissions

    def _set_tables(self, tables):
        self._tables = tables

    def _set_default_permissions(self, permissions, content_type):
        attr = "_default_{content}_permissions".format(content=content_type)
        setattr(
            self,
            attr,
            permissions,
        )
        logging.getLogger().debug({"type": attr, "value": getattr(self, attr)})

    def _set_data_quality_warnings(self, dqw):
        self._data_quality_warnings = dqw

    @classmethod
    def from_response(cls, resp, ns):
        all_database_items = list()
        parsed_response = fromstring(resp)
        all_database_xml = parsed_response.findall(".//t:database", namespaces=ns)

        for database_xml in all_database_xml:
            parsed_database = cls._parse_element(database_xml, ns)
            database_item = cls(parsed_database["name"])
            database_item._set_values(parsed_database)
            all_database_items.append(database_item)
        return all_database_items

    @staticmethod
    def _parse_element(database_xml, ns):
        database_values = database_xml.attrib.copy()
        contact = database_xml.find(".//t:contact", namespaces=ns)
        if contact is not None:
            database_values["contact"] = contact.attrib.copy()
        return database_values


# Used to convert string represented boolean to a boolean type
def string_to_bool(s):
    return s.lower() == "true"



import os
import unittest

import requests_mock

import tableauserverclient as TSC
from tableauserverclient.datetime_helpers import format_datetime

TEST_ASSET_DIR = os.path.join(os.path.dirname(__file__), "assets")

GET_XML = os.path.join(TEST_ASSET_DIR, "custom_view_get.xml")
GET_XML_ID = os.path.join(TEST_ASSET_DIR, "custom_view_get_id.xml")
POPULATE_PREVIEW_IMAGE = os.path.join(TEST_ASSET_DIR, "Sample View Image.png")
CUSTOM_VIEW_UPDATE_XML = os.path.join(TEST_ASSET_DIR, "custom_view_update.xml")


class CustomViewTests(unittest.TestCase):
    def setUp(self):
        self.server = TSC.Server("http://test", False)
        self.server.version = "3.19"  # custom views only introduced in 3.19

        # Fake sign in
        self.server._site_id = "dad65087-b08b-4603-af4e-2887b8aafc67"
        self.server._auth_token = "j80k54ll2lfMZ0tv97mlPvvSCRyD0DOM"

        self.baseurl = self.server.custom_views.baseurl

    def test_get(self) -> None:
        with open(GET_XML, "rb") as f:
            response_xml = f.read().decode("utf-8")
            print(response_xml)
        with requests_mock.mock() as m:
            m.get(self.baseurl, text=response_xml)
            all_views, pagination_item = self.server.custom_views.get()

        self.assertEqual(2, pagination_item.total_available)
        self.assertEqual("d79634e1-6063-4ec9-95ff-50acbf609ff5", all_views[0].id)
        self.assertEqual("ENDANGERED SAFARI", all_views[0].name)
        self.assertEqual("SafariSample/sheets/ENDANGEREDSAFARI", all_views[0].content_url)
        self.assertEqual("3cc6cd06-89ce-4fdc-b935-5294135d6d42", all_views[0].workbook.id)
        self.assertEqual("5de011f8-5aa9-4d5b-b991-f462c8dd6bb7", all_views[0].owner.id)
        self.assertIsNone(all_views[0].created_at)
        self.assertIsNone(all_views[0].updated_at)

        self.assertEqual("fd252f73-593c-4c4e-8584-c032b8022adc", all_views[1].id)
        self.assertEqual("Overview", all_views[1].name)
        self.assertEqual(False, all_views[1].shared)
        self.assertEqual("6d13b0ca-043d-4d42-8c9d-3f3313ea3a00", all_views[1].workbook.id)
        self.assertEqual("5de011f8-5aa9-4d5b-b991-f462c8dd6bb7", all_views[1].owner.id)
        self.assertEqual("2002-05-30T09:00:00Z", format_datetime(all_views[1].created_at))
        self.assertEqual("2002-06-05T08:00:59Z", format_datetime(all_views[1].updated_at))

    def test_get_by_id(self) -> None:
        with open(GET_XML_ID, "rb") as f:
            response_xml = f.read().decode("utf-8")
        with requests_mock.mock() as m:
            m.get(self.baseurl + "/d79634e1-6063-4ec9-95ff-50acbf609ff5", text=response_xml)
            view: TSC.CustomViewItem = self.server.custom_views.get_by_id("d79634e1-6063-4ec9-95ff-50acbf609ff5")

        self.assertEqual("d79634e1-6063-4ec9-95ff-50acbf609ff5", view.id)
        self.assertEqual("ENDANGERED SAFARI", view.name)
        self.assertEqual("SafariSample/sheets/ENDANGEREDSAFARI", view.content_url)
        if view.workbook:
            self.assertEqual("3cc6cd06-89ce-4fdc-b935-5294135d6d42", view.workbook.id)
        if view.owner:
            self.assertEqual("5de011f8-5aa9-4d5b-b991-f462c8dd6bb7", view.owner.id)
        if view.view:
            self.assertEqual("5241e88d-d384-4fd7-9c2f-648b5247efc5", view.view.id)
        self.assertEqual("2002-05-30T09:00:00Z", format_datetime(view.created_at))
        self.assertEqual("2002-06-05T08:00:59Z", format_datetime(view.updated_at))

    def test_get_by_id_missing_id(self) -> None:
        self.assertRaises(TSC.MissingRequiredFieldError, self.server.custom_views.get_by_id, None)

    def test_get_before_signin(self) -> None:
        self.server._auth_token = None
        self.assertRaises(TSC.NotSignedInError, self.server.custom_views.get)

    def test_populate_image(self) -> None:
        with open(POPULATE_PREVIEW_IMAGE, "rb") as f:
            response = f.read()
        with requests_mock.mock() as m:
            m.get(self.baseurl + "/d79634e1-6063-4ec9-95ff-50acbf609ff5/image", content=response)
            single_view = TSC.CustomViewItem()
            single_view._id = "d79634e1-6063-4ec9-95ff-50acbf609ff5"
            self.server.custom_views.populate_image(single_view)
            self.assertEqual(response, single_view.image)

    def test_populate_image_with_options(self) -> None:
        with open(POPULATE_PREVIEW_IMAGE, "rb") as f:
            response = f.read()
        with requests_mock.mock() as m:
            m.get(
                self.baseurl + "/d79634e1-6063-4ec9-95ff-50acbf609ff5/image?resolution=high&maxAge=10", content=response
            )
            single_view = TSC.CustomViewItem()
            single_view._id = "d79634e1-6063-4ec9-95ff-50acbf609ff5"
            req_option = TSC.ImageRequestOptions(imageresolution=TSC.ImageRequestOptions.Resolution.High, maxage=10)
            self.server.custom_views.populate_image(single_view, req_option)
            self.assertEqual(response, single_view.image)

    def test_populate_image_missing_id(self) -> None:
        single_view = TSC.CustomViewItem()
        single_view._id = None
        self.assertRaises(TSC.MissingRequiredFieldError, self.server.custom_views.populate_image, single_view)

    def test_delete(self) -> None:
        with requests_mock.mock() as m:
            m.delete(self.baseurl + "/3cc6cd06-89ce-4fdc-b935-5294135d6d42", status_code=204)
            self.server.custom_views.delete("3cc6cd06-89ce-4fdc-b935-5294135d6d42")

    def test_delete_missing_id(self) -> None:
        self.assertRaises(ValueError, self.server.custom_views.delete, "")

    def test_update(self) -> None:
        with open(CUSTOM_VIEW_UPDATE_XML, "rb") as f:
            response_xml = f.read().decode("utf-8")
        with requests_mock.mock() as m:
            m.put(self.baseurl + "/1f951daf-4061-451a-9df1-69a8062664f2", text=response_xml)
            the_custom_view = TSC.CustomViewItem("1d0304cd-3796-429f-b815-7258370b9b74", name="Best test ever")
            the_custom_view._id = "1f951daf-4061-451a-9df1-69a8062664f2"
            the_custom_view.owner = TSC.UserItem()
            the_custom_view.owner.id = "dd2239f6-ddf1-4107-981a-4cf94e415794"
            the_custom_view = self.server.custom_views.update(the_custom_view)

        self.assertEqual("1f951daf-4061-451a-9df1-69a8062664f2", the_custom_view.id)
        if the_custom_view.owner:
            self.assertEqual("dd2239f6-ddf1-4107-981a-4cf94e415794", the_custom_view.owner.id)
        self.assertEqual("Best test ever", the_custom_view.name)

    def test_update_missing_id(self) -> None:
        cv = TSC.CustomViewItem(name="test")
        self.assertRaises(TSC.MissingRequiredFieldError, self.server.custom_views.update, cv)



from typing import Tuple
from .filter import Filter
from .request_options import RequestOptions
from .sort import Sort
import math


def to_camel_case(word: str) -> str:
    return word.split("_")[0] + "".join(x.capitalize() or "_" for x in word.split("_")[1:])


"""
This interface allows more fluent queries against Tableau Server
e.g server.users.get(name="user@domain.com")
see pagination_sample
"""


class QuerySet:
    def __init__(self, model):
        self.model = model
        self.request_options = RequestOptions()
        self._result_cache = None
        self._pagination_item = None

    def __iter__(self):
        # Not built to be re-entrant. Starts back at page 1, and empties
        # the result cache.
        self.request_options.pagenumber = 1
        self._result_cache = None
        total = self.total_available
        size = self.page_size
        yield from self._result_cache

        # Loop through the subsequent pages.
        for page in range(1, math.ceil(total / size)):
            self.request_options.pagenumber = page + 1
            self._result_cache = None
            self._fetch_all()
            yield from self._result_cache

    def __getitem__(self, k):
        page = self.page_number
        size = self.page_size

        # Create a range object for quick checking if k is in the cached result.
        page_range = range((page - 1) * size, page * size)

        if isinstance(k, slice):
            # Parse out the slice object, and assume reasonable defaults if no value provided.
            step = k.step if k.step is not None else 1
            start = k.start if k.start is not None else 0
            stop = k.stop if k.stop is not None else self.total_available

            # If negative values present in slice, convert to positive values
            if start < 0:
                start += self.total_available
            if stop < 0:
                stop += self.total_available
            if start < stop and step < 0:
                # Since slicing is left inclusive and right exclusive, shift
                # the start and stop values by 1 to keep that behavior
                start, stop = stop - 1, start - 1
                slice_stop = stop if stop > 0 else None
                k = slice(start, slice_stop, step)

            # Fetch items from cache if present, otherwise, recursively fetch.
            k_range = range(start, stop, step)
            if all(i in page_range for i in k_range):
                return self._result_cache[k]
            return [self[i] for i in k_range]

        if k < 0:
            k += self.total_available

        if k in page_range:
            # Fetch item from cache if present
            return self._result_cache[k % size]
        elif k in range(self.total_available):
            # Otherwise, check if k is even sensible to return
            self._result_cache = None
            # Add one to k, otherwise it gets stuck at page boundaries, e.g. 100
            self.request_options.pagenumber = max(1, math.ceil((k + 1) / size))
            return self[k]
        else:
            # If k is unreasonable, raise an IndexError.
            raise IndexError

    def _fetch_all(self):
        """
        Retrieve the data and store result and pagination item in cache
        """
        if self._result_cache is None:
            self._result_cache, self._pagination_item = self.model.get(self.request_options)

    def __len__(self) -> int:
        return self.total_available

    @property
    def total_available(self) -> int:
        self._fetch_all()
        return self._pagination_item.total_available

    @property
    def page_number(self) -> int:
        self._fetch_all()
        return self._pagination_item.page_number

    @property
    def page_size(self) -> int:
        self._fetch_all()
        return self._pagination_item.page_size

    def filter(self, *invalid, **kwargs):
        if invalid:
            raise RuntimeError(f"Only accepts keyword arguments.")
        for kwarg_key, value in kwargs.items():
            field_name, operator = self._parse_shorthand_filter(kwarg_key)
            self.request_options.filter.add(Filter(field_name, operator, value))
        return self

    def order_by(self, *args):
        for arg in args:
            field_name, direction = self._parse_shorthand_sort(arg)
            self.request_options.sort.add(Sort(field_name, direction))
        return self

    def paginate(self, **kwargs):
        if "page_number" in kwargs:
            self.request_options.pagenumber = kwargs["page_number"]
        if "page_size" in kwargs:
            self.request_options.pagesize = kwargs["page_size"]
        return self

    def _parse_shorthand_filter(self, key: str) -> Tuple[str, str]:
        tokens = key.split("__", 1)
        if len(tokens) == 1:
            operator = RequestOptions.Operator.Equals
        else:
            operator = tokens[1]
            if operator not in RequestOptions.Operator.__dict__.values():
                raise ValueError("Operator `{}` is not valid.".format(operator))

        field = to_camel_case(tokens[0])
        if field not in RequestOptions.Field.__dict__.values():
            raise ValueError("Field name `{}` is not valid.".format(field))
        return (field, operator)

    def _parse_shorthand_sort(self, key: str) -> Tuple[str, str]:
        direction = RequestOptions.Direction.Asc
        if key.startswith("-"):
            direction = RequestOptions.Direction.Desc
            key = key[1:]

        key = to_camel_case(key)
        if key not in RequestOptions.Field.__dict__.values():
            raise ValueError("Sort key name %s is not valid.", key)
        return (key, direction)



import cgi
import copy
import json
import logging
import io
import os

from contextlib import closing
from pathlib import Path
from typing import List, Mapping, Optional, Sequence, Tuple, TYPE_CHECKING, Union

if TYPE_CHECKING:
    from tableauserverclient.server import Server
    from tableauserverclient.models import PermissionsRule
    from .schedules_endpoint import AddResponse

from .dqw_endpoint import _DataQualityWarningEndpoint
from .endpoint import QuerysetEndpoint, api, parameter_added_in
from .exceptions import InternalServerError, MissingRequiredFieldError
from .permissions_endpoint import _PermissionsEndpoint
from .resource_tagger import _ResourceTagger

from tableauserverclient.server import RequestFactory, RequestOptions
from tableauserverclient.filesys_helpers import (
    to_filename,
    make_download_path,
    get_file_type,
    get_file_object_size,
)
from tableauserverclient.models import (
    ConnectionCredentials,
    ConnectionItem,
    DatasourceItem,
    JobItem,
    RevisionItem,
    PaginationItem,
)

io_types = (io.BytesIO, io.BufferedReader)
io_types_r = (io.BytesIO, io.BufferedReader)
io_types_w = (io.BytesIO, io.BufferedWriter)

FilePath = Union[str, os.PathLike]
FileObject = Union[io.BufferedReader, io.BytesIO]
PathOrFile = Union[FilePath, FileObject]

# The maximum size of a file that can be published in a single request is 64MB
FILESIZE_LIMIT = 1024 * 1024 * 64  # 64MB

ALLOWED_FILE_EXTENSIONS = ["tds", "tdsx", "tde", "hyper", "parquet"]

logger = logging.getLogger("tableau.endpoint.datasources")

FilePath = Union[str, os.PathLike]
FileObjectR = Union[io.BufferedReader, io.BytesIO]
FileObjectW = Union[io.BufferedWriter, io.BytesIO]
PathOrFileR = Union[FilePath, FileObjectR]
PathOrFileW = Union[FilePath, FileObjectW]


class Datasources(QuerysetEndpoint):
    def __init__(self, parent_srv: "Server") -> None:
        super(Datasources, self).__init__(parent_srv)
        self._resource_tagger = _ResourceTagger(parent_srv)
        self._permissions = _PermissionsEndpoint(parent_srv, lambda: self.baseurl)
        self._data_quality_warnings = _DataQualityWarningEndpoint(self.parent_srv, "datasource")

        return None

    @property
    def baseurl(self) -> str:
        return "{0}/sites/{1}/datasources".format(self.parent_srv.baseurl, self.parent_srv.site_id)

    # Get all datasources
    @api(version="2.0")
    def get(self, req_options: Optional[RequestOptions] = None) -> Tuple[List[DatasourceItem], PaginationItem]:
        logger.info("Querying all datasources on site")
        url = self.baseurl
        server_response = self.get_request(url, req_options)
        pagination_item = PaginationItem.from_response(server_response.content, self.parent_srv.namespace)
        all_datasource_items = DatasourceItem.from_response(server_response.content, self.parent_srv.namespace)
        return all_datasource_items, pagination_item

    # Get 1 datasource by id
    @api(version="2.0")
    def get_by_id(self, datasource_id: str) -> DatasourceItem:
        if not datasource_id:
            error = "Datasource ID undefined."
            raise ValueError(error)
        logger.info("Querying single datasource (ID: {0})".format(datasource_id))
        url = "{0}/{1}".format(self.baseurl, datasource_id)
        server_response = self.get_request(url)
        return DatasourceItem.from_response(server_response.content, self.parent_srv.namespace)[0]

    # Populate datasource item's connections
    @api(version="2.0")
    def populate_connections(self, datasource_item: DatasourceItem) -> None:
        if not datasource_item.id:
            error = "Datasource item missing ID. Datasource must be retrieved from server first."
            raise MissingRequiredFieldError(error)

        def connections_fetcher():
            return self._get_datasource_connections(datasource_item)

        datasource_item._set_connections(connections_fetcher)
        logger.info("Populated connections for datasource (ID: {0})".format(datasource_item.id))

    def _get_datasource_connections(self, datasource_item, req_options=None):
        url = "{0}/{1}/connections".format(self.baseurl, datasource_item.id)
        server_response = self.get_request(url, req_options)
        connections = ConnectionItem.from_response(server_response.content, self.parent_srv.namespace)
        return connections

    # Delete 1 datasource by id
    @api(version="2.0")
    def delete(self, datasource_id: str) -> None:
        if not datasource_id:
            error = "Datasource ID undefined."
            raise ValueError(error)
        url = "{0}/{1}".format(self.baseurl, datasource_id)
        self.delete_request(url)
        logger.info("Deleted single datasource (ID: {0})".format(datasource_id))

    # Download 1 datasource by id
    @api(version="2.0")
    @parameter_added_in(no_extract="2.5")
    @parameter_added_in(include_extract="2.5")
    def download(
        self,
        datasource_id: str,
        filepath: Optional[PathOrFileW] = None,
        include_extract: bool = True,
        no_extract: Optional[bool] = None,
    ) -> str:
        return self.download_revision(datasource_id, None, filepath, include_extract, no_extract)

    # Update datasource
    @api(version="2.0")
    def update(self, datasource_item: DatasourceItem) -> DatasourceItem:
        if not datasource_item.id:
            error = "Datasource item missing ID. Datasource must be retrieved from server first."
            raise MissingRequiredFieldError(error)
        # bug - before v3.15 you must always include the project id
        if datasource_item.owner_id and not datasource_item.project_id:
            if not self.parent_srv.check_at_least_version("3.15"):
                error = (
                    "Attempting to set new owner but datasource is missing Project ID."
                    "In versions before 3.15 the project id must be included to update the owner."
                )
                raise MissingRequiredFieldError(error)

        self._resource_tagger.update_tags(self.baseurl, datasource_item)

        # Update the datasource itself
        url = "{0}/{1}".format(self.baseurl, datasource_item.id)

        update_req = RequestFactory.Datasource.update_req(datasource_item)
        server_response = self.put_request(url, update_req)
        logger.info("Updated datasource item (ID: {0})".format(datasource_item.id))
        updated_datasource = copy.copy(datasource_item)
        return updated_datasource._parse_common_elements(server_response.content, self.parent_srv.namespace)

    # Update datasource connections
    @api(version="2.3")
    def update_connection(self, datasource_item: DatasourceItem, connection_item: ConnectionItem) -> ConnectionItem:
        url = "{0}/{1}/connections/{2}".format(self.baseurl, datasource_item.id, connection_item.id)

        update_req = RequestFactory.Connection.update_req(connection_item)
        server_response = self.put_request(url, update_req)
        connection = ConnectionItem.from_response(server_response.content, self.parent_srv.namespace)[0]

        logger.info(
            "Updated datasource item (ID: {0} & connection item {1}".format(datasource_item.id, connection_item.id)
        )
        return connection

    @api(version="2.8")
    def refresh(self, datasource_item: DatasourceItem) -> JobItem:
        id_ = getattr(datasource_item, "id", datasource_item)
        url = "{0}/{1}/refresh".format(self.baseurl, id_)
        empty_req = RequestFactory.Empty.empty_req()
        server_response = self.post_request(url, empty_req)
        new_job = JobItem.from_response(server_response.content, self.parent_srv.namespace)[0]
        return new_job

    @api(version="3.5")
    def create_extract(self, datasource_item: DatasourceItem, encrypt: bool = False) -> JobItem:
        id_ = getattr(datasource_item, "id", datasource_item)
        url = "{0}/{1}/createExtract?encrypt={2}".format(self.baseurl, id_, encrypt)
        empty_req = RequestFactory.Empty.empty_req()
        server_response = self.post_request(url, empty_req)
        new_job = JobItem.from_response(server_response.content, self.parent_srv.namespace)[0]
        return new_job

    @api(version="3.5")
    def delete_extract(self, datasource_item: DatasourceItem) -> None:
        id_ = getattr(datasource_item, "id", datasource_item)
        url = "{0}/{1}/deleteExtract".format(self.baseurl, id_)
        empty_req = RequestFactory.Empty.empty_req()
        self.post_request(url, empty_req)

    # Publish datasource
    @api(version="2.0")
    @parameter_added_in(connections="2.8")
    @parameter_added_in(as_job="3.0")
    def publish(
        self,
        datasource_item: DatasourceItem,
        file: PathOrFileR,
        mode: str,
        connection_credentials: Optional[ConnectionCredentials] = None,
        connections: Optional[Sequence[ConnectionItem]] = None,
        as_job: bool = False,
    ) -> Union[DatasourceItem, JobItem]:
        if isinstance(file, (os.PathLike, str)):
            if not os.path.isfile(file):
                error = "File path does not lead to an existing file."
                raise IOError(error)

            filename = os.path.basename(file)
            file_extension = os.path.splitext(filename)[1][1:]
            file_size = os.path.getsize(file)

            # If name is not defined, grab the name from the file to publish
            if not datasource_item.name:
                datasource_item.name = os.path.splitext(filename)[0]
            if file_extension not in ALLOWED_FILE_EXTENSIONS:
                error = "Only {} files can be published as datasources.".format(", ".join(ALLOWED_FILE_EXTENSIONS))
                raise ValueError(error)

        elif isinstance(file, io_types_r):
            if not datasource_item.name:
                error = "Datasource item must have a name when passing a file object"
                raise ValueError(error)

            file_type = get_file_type(file)
            if file_type == "zip":
                file_extension = "tdsx"
            elif file_type == "xml":
                file_extension = "tds"
            else:
                error = "Unsupported file type {}".format(file_type)
                raise ValueError(error)

            filename = "{}.{}".format(datasource_item.name, file_extension)
            file_size = get_file_object_size(file)

        else:
            raise TypeError("file should be a filepath or file object.")

        if not mode or not hasattr(self.parent_srv.PublishMode, mode):
            error = "Invalid mode defined."
            raise ValueError(error)

        # Construct the url with the defined mode
        url = "{0}?datasourceType={1}".format(self.baseurl, file_extension)
        if mode == self.parent_srv.PublishMode.Overwrite or mode == self.parent_srv.PublishMode.Append:
            url += "&{0}=true".format(mode.lower())

        if as_job:
            url += "&{0}=true".format("asJob")

        # Determine if chunking is required (64MB is the limit for single upload method)
        if file_size >= FILESIZE_LIMIT:
            logger.info("Publishing {0} to server with chunking method (datasource over 64MB)".format(filename))
            upload_session_id = self.parent_srv.fileuploads.upload(file)
            url = "{0}&uploadSessionId={1}".format(url, upload_session_id)
            xml_request, content_type = RequestFactory.Datasource.publish_req_chunked(
                datasource_item, connection_credentials, connections
            )
        else:
            logger.info("Publishing {0} to server".format(filename))

            if isinstance(file, (Path, str)):
                with open(file, "rb") as f:
                    file_contents = f.read()
            elif isinstance(file, io_types_r):
                file_contents = file.read()
            else:
                raise TypeError("file should be a filepath or file object.")

            xml_request, content_type = RequestFactory.Datasource.publish_req(
                datasource_item,
                filename,
                file_contents,
                connection_credentials,
                connections,
            )

        # Send the publishing request to server
        try:
            server_response = self.post_request(url, xml_request, content_type)
        except InternalServerError as err:
            if err.code == 504 and not as_job:
                err.content = "Timeout error while publishing. Please use asynchronous publishing to avoid timeouts."
            raise err

        if as_job:
            new_job = JobItem.from_response(server_response.content, self.parent_srv.namespace)[0]
            logger.info("Published {0} (JOB_ID: {1}".format(filename, new_job.id))
            return new_job
        else:
            new_datasource = DatasourceItem.from_response(server_response.content, self.parent_srv.namespace)[0]
            logger.info("Published {0} (ID: {1})".format(filename, new_datasource.id))
            return new_datasource

    @api(version="3.13")
    def update_hyper_data(
        self,
        datasource_or_connection_item: Union[DatasourceItem, ConnectionItem, str],
        *,
        request_id: str,
        actions: Sequence[Mapping],
        payload: Optional[FilePath] = None,
    ) -> JobItem:
        if isinstance(datasource_or_connection_item, DatasourceItem):
            datasource_id = datasource_or_connection_item.id
            url = "{0}/{1}/data".format(self.baseurl, datasource_id)
        elif isinstance(datasource_or_connection_item, ConnectionItem):
            datasource_id = datasource_or_connection_item.datasource_id
            connection_id = datasource_or_connection_item.id
            url = "{0}/{1}/connections/{2}/data".format(self.baseurl, datasource_id, connection_id)
        else:
            assert isinstance(datasource_or_connection_item, str)
            url = "{0}/{1}/data".format(self.baseurl, datasource_or_connection_item)

        if payload is not None:
            if not os.path.isfile(payload):
                error = "File path does not lead to an existing file."
                raise IOError(error)

            logger.info("Uploading {0} to server with chunking method for Update job".format(payload))
            upload_session_id = self.parent_srv.fileuploads.upload(payload)
            url = "{0}?uploadSessionId={1}".format(url, upload_session_id)

        json_request = json.dumps({"actions": actions})
        parameters = {"headers": {"requestid": request_id}}
        server_response = self.patch_request(url, json_request, "application/json", parameters=parameters)
        new_job = JobItem.from_response(server_response.content, self.parent_srv.namespace)[0]
        return new_job

    @api(version="2.0")
    def populate_permissions(self, item: DatasourceItem) -> None:
        self._permissions.populate(item)

    @api(version="2.0")
    def update_permission(self, item, permission_item):
        import warnings

        warnings.warn(
            "Server.datasources.update_permission is deprecated, "
            "please use Server.datasources.update_permissions instead.",
            DeprecationWarning,
        )
        self._permissions.update(item, permission_item)

    @api(version="2.0")
    def update_permissions(self, item: DatasourceItem, permission_item: List["PermissionsRule"]) -> None:
        self._permissions.update(item, permission_item)

    @api(version="2.0")
    def delete_permission(self, item: DatasourceItem, capability_item: "PermissionsRule") -> None:
        self._permissions.delete(item, capability_item)

    @api(version="3.5")
    def populate_dqw(self, item):
        self._data_quality_warnings.populate(item)

    @api(version="3.5")
    def update_dqw(self, item, warning):
        return self._data_quality_warnings.update(item, warning)

    @api(version="3.5")
    def add_dqw(self, item, warning):
        return self._data_quality_warnings.add(item, warning)

    @api(version="3.5")
    def delete_dqw(self, item):
        self._data_quality_warnings.clear(item)

    # Populate datasource item's revisions
    @api(version="2.3")
    def populate_revisions(self, datasource_item: DatasourceItem) -> None:
        if not datasource_item.id:
            error = "Datasource item missing ID. Datasource must be retrieved from server first."
            raise MissingRequiredFieldError(error)

        def revisions_fetcher():
            return self._get_datasource_revisions(datasource_item)

        datasource_item._set_revisions(revisions_fetcher)
        logger.info("Populated revisions for datasource (ID: {0})".format(datasource_item.id))

    def _get_datasource_revisions(
        self, datasource_item: DatasourceItem, req_options: Optional["RequestOptions"] = None
    ) -> List[RevisionItem]:
        url = "{0}/{1}/revisions".format(self.baseurl, datasource_item.id)
        server_response = self.get_request(url, req_options)
        revisions = RevisionItem.from_response(server_response.content, self.parent_srv.namespace, datasource_item)
        return revisions

    # Download 1 datasource revision by revision number
    @api(version="2.3")
    def download_revision(
        self,
        datasource_id: str,
        revision_number: str,
        filepath: Optional[PathOrFileW] = None,
        include_extract: bool = True,
        no_extract: Optional[bool] = None,
    ) -> PathOrFileW:
        if not datasource_id:
            error = "Datasource ID undefined."
            raise ValueError(error)
        if revision_number is None:
            url = "{0}/{1}/content".format(self.baseurl, datasource_id)
        else:
            url = "{0}/{1}/revisions/{2}/content".format(self.baseurl, datasource_id, revision_number)
        if no_extract is False or no_extract is True:
            import warnings

            warnings.warn(
                "no_extract is deprecated, use include_extract instead.",
                DeprecationWarning,
            )
            include_extract = not no_extract

        if not include_extract:
            url += "?includeExtract=False"

        with closing(self.get_request(url, parameters={"stream": True})) as server_response:
            _, params = cgi.parse_header(server_response.headers["Content-Disposition"])
            if isinstance(filepath, io_types_w):
                for chunk in server_response.iter_content(1024):  # 1KB
                    filepath.write(chunk)
                return_path = filepath
            else:
                filename = to_filename(os.path.basename(params["filename"]))
                download_path = make_download_path(filepath, filename)
                with open(download_path, "wb") as f:
                    for chunk in server_response.iter_content(1024):  # 1KB
                        f.write(chunk)
                return_path = os.path.abspath(download_path)

        logger.info(
            "Downloaded datasource revision {0} to {1} (ID: {2})".format(revision_number, return_path, datasource_id)
        )
        return return_path

    @api(version="2.3")
    def delete_revision(self, datasource_id: str, revision_number: str) -> None:
        if datasource_id is None or revision_number is None:
            raise ValueError
        url = "/".join([self.baseurl, datasource_id, "revisions", revision_number])

        self.delete_request(url)
        logger.info(
            "Deleted single datasource revision (ID: {0}) (Revision: {1})".format(datasource_id, revision_number)
        )

    # a convenience method
    @api(version="2.8")
    def schedule_extract_refresh(
        self, schedule_id: str, item: DatasourceItem
    ) -> List["AddResponse"]:  # actually should return a task
        return self.parent_srv.schedules.add_to_schedule(schedule_id, datasource=item)



from tableauserverclient.models.property_decorators import property_is_int
import logging

logger = logging.getLogger("tableau.request_options")


class RequestOptionsBase(object):
    # This method is used if server api version is below 3.7 (2020.1)
    def apply_query_params(self, url):
        try:
            params = self.get_query_params()
            params_list = ["{}={}".format(k, v) for (k, v) in params.items()]

            logger.debug("Applying options to request: <%s(%s)>", self.__class__.__name__, ",".join(params_list))

            if "?" in url:
                url, existing_params = url.split("?")
                params_list.append(existing_params)

            return "{0}?{1}".format(url, "&".join(params_list))
        except NotImplementedError:
            raise

    def get_query_params(self):
        raise NotImplementedError()


class RequestOptions(RequestOptionsBase):
    class Operator:
        Equals = "eq"
        GreaterThan = "gt"
        GreaterThanOrEqual = "gte"
        LessThan = "lt"
        LessThanOrEqual = "lte"
        In = "in"
        Has = "has"

    class Field:
        Args = "args"
        CompletedAt = "completedAt"
        CreatedAt = "createdAt"
        DomainName = "domainName"
        DomainNickname = "domainNickname"
        HitsTotal = "hitsTotal"
        IsLocal = "isLocal"
        JobType = "jobType"
        LastLogin = "lastLogin"
        MinimumSiteRole = "minimumSiteRole"
        Name = "name"
        Notes = "notes"
        OwnerDomain = "ownerDomain"
        OwnerEmail = "ownerEmail"
        OwnerName = "ownerName"
        ParentProjectId = "parentProjectId"
        Progress = "progress"
        ProjectName = "projectName"
        PublishSamples = "publishSamples"
        SiteRole = "siteRole"
        StartedAt = "startedAt"
        Status = "status"
        Subtitle = "subtitle"
        Tags = "tags"
        Title = "title"
        TopLevelProject = "topLevelProject"
        Type = "type"
        UpdatedAt = "updatedAt"
        UserCount = "userCount"

    class Direction:
        Desc = "desc"
        Asc = "asc"

    def __init__(self, pagenumber=1, pagesize=100):
        self.pagenumber = pagenumber
        self.pagesize = pagesize
        self.sort = set()
        self.filter = set()

        # This is private until we expand all of our parsers to handle the extra fields
        self._all_fields = False

    def page_size(self, page_size):
        self.pagesize = page_size
        return self

    def page_number(self, page_number):
        self.pagenumber = page_number
        return self

    def get_query_params(self):
        params = {}
        if self.pagenumber:
            params["pageNumber"] = self.pagenumber
        if self.pagesize:
            params["pageSize"] = self.pagesize
        if len(self.sort) > 0:
            sort_options = (str(sort_item) for sort_item in self.sort)
            ordered_sort_options = sorted(sort_options)
            params["sort"] = ",".join(ordered_sort_options)
        if len(self.filter) > 0:
            filter_options = (str(filter_item) for filter_item in self.filter)
            ordered_filter_options = sorted(filter_options)
            params["filter"] = ",".join(ordered_filter_options)
        if self._all_fields:
            params["fields"] = "_all_"
        return params


class _FilterOptionsBase(RequestOptionsBase):
    """Provide a basic implementation of adding view filters to the url"""

    def __init__(self):
        self.view_filters = []

    def get_query_params(self):
        raise NotImplementedError()

    def vf(self, name, value):
        self.view_filters.append((name, value))
        return self

    def _append_view_filters(self, params):
        for name, value in self.view_filters:
            params["vf_" + name] = value


class CSVRequestOptions(_FilterOptionsBase):
    def __init__(self, maxage=-1):
        super(CSVRequestOptions, self).__init__()
        self.max_age = maxage

    @property
    def max_age(self):
        return self._max_age

    @max_age.setter
    @property_is_int(range=(0, 240), allowed=[-1])
    def max_age(self, value):
        self._max_age = value

    def get_query_params(self):
        params = {}
        if self.max_age != -1:
            params["maxAge"] = self.max_age

        self._append_view_filters(params)
        return params


class ExcelRequestOptions(RequestOptionsBase):
    def __init__(self, maxage: int = -1) -> None:
        super().__init__()
        self.max_age = maxage

    @property
    def max_age(self) -> int:
        return self._max_age

    @max_age.setter
    @property_is_int(range=(0, 240), allowed=[-1])
    def max_age(self, value: int) -> None:
        self._max_age = value

    def get_query_params(self):
        params = {}
        if self.max_age != -1:
            params["maxAge"] = self.max_age

        return params


class ImageRequestOptions(_FilterOptionsBase):
    # if 'high' isn't specified, the REST API endpoint returns an image with standard resolution
    class Resolution:
        High = "high"

    def __init__(self, imageresolution=None, maxage=-1):
        super(ImageRequestOptions, self).__init__()
        self.image_resolution = imageresolution
        self.max_age = maxage

    @property
    def max_age(self):
        return self._max_age

    @max_age.setter
    @property_is_int(range=(0, 240), allowed=[-1])
    def max_age(self, value):
        self._max_age = value

    def get_query_params(self):
        params = {}
        if self.image_resolution:
            params["resolution"] = self.image_resolution
        if self.max_age != -1:
            params["maxAge"] = self.max_age
        self._append_view_filters(params)
        return params


class PDFRequestOptions(_FilterOptionsBase):
    class PageType:
        A3 = "a3"
        A4 = "a4"
        A5 = "a5"
        B4 = "b4"
        B5 = "b5"
        Executive = "executive"
        Folio = "folio"
        Ledger = "ledger"
        Legal = "legal"
        Letter = "letter"
        Note = "note"
        Quarto = "quarto"
        Tabloid = "tabloid"
        Unspecified = "unspecified"

    class Orientation:
        Portrait = "portrait"
        Landscape = "landscape"

    def __init__(self, page_type=None, orientation=None, maxage=-1):
        super(PDFRequestOptions, self).__init__()
        self.page_type = page_type
        self.orientation = orientation
        self.max_age = maxage

    @property
    def max_age(self):
        return self._max_age

    @max_age.setter
    @property_is_int(range=(0, 240), allowed=[-1])
    def max_age(self, value):
        self._max_age = value

    def get_query_params(self):
        params = {}
        if self.page_type:
            params["type"] = self.page_type

        if self.orientation:
            params["orientation"] = self.orientation

        if self.max_age != -1:
            params["maxAge"] = self.max_age

        self._append_view_filters(params)

        return params



####
# This script demonstrates how to create a group using the Tableau
# Server Client.
#
# To run the script, you must have installed Python 3.7 or later.
####


import argparse
import logging
import os

from datetime import time
from typing import List

import tableauserverclient as TSC
from tableauserverclient import ServerResponseError


def main():
    parser = argparse.ArgumentParser(description="Creates a sample user group.")
    # Common options; please keep those in sync across all samples
    parser.add_argument("--server", "-s", required=True, help="server address")
    parser.add_argument("--site", "-S", help="site name")
    parser.add_argument(
        "--token-name", "-p", required=True, help="name of the personal access token used to sign into the server"
    )
    parser.add_argument(
        "--token-value", "-v", required=True, help="value of the personal access token used to sign into the server"
    )
    parser.add_argument(
        "--logging-level",
        "-l",
        choices=["debug", "info", "error"],
        default="error",
        help="desired logging level (set to error by default)",
    )
    # Options specific to this sample
    # This sample has no additional options, yet. If you add some, please add them here
    parser.add_argument("--file", help="csv file containing user info", required=False)
    args = parser.parse_args()

    # Set logging level based on user input, or error by default
    logging_level = getattr(logging, args.logging_level.upper())
    logging.basicConfig(level=logging_level)

    tableau_auth = TSC.PersonalAccessTokenAuth(args.token_name, args.token_value, site_id=args.site)
    server = TSC.Server(args.server, use_server_version=True, http_options={"verify": False})
    with server.auth.sign_in(tableau_auth):
        # this code shows 3 different error codes that mean "resource is already in collection"
        # 409009: group already exists on server
        # 409107: user is already on site
        # 409011: user is already in group

        group = TSC.GroupItem("test")
        try:
            group = server.groups.create(group)
        except TSC.server.endpoint.exceptions.ServerResponseError as rError:
            if rError.code == "409009":
                print("Group already exists")
                group = server.groups.filter(name=group.name)[0]
            else:
                raise rError
        server.groups.populate_users(group)
        for user in group.users:
            print(user.name)

        if args.file:
            filepath = os.path.abspath(args.file)
            print("Add users to site from file {}:".format(filepath))
            added: List[TSC.UserItem]
            failed: List[TSC.UserItem, TSC.ServerResponseError]
            added, failed = server.users.create_from_file(filepath)
            for user, error in failed:
                print(user, error.code)
                if error.code == "409017":
                    user = server.users.filter(name=user.name)[0]
                    added.append(user)
            print("Adding users to group:{}".format(added))
            for user in added:
                print("Adding user {}".format(user))
                try:
                    server.groups.add_user(group, user.id)
                except ServerResponseError as serverError:
                    if serverError.code == "409011":
                        print("user {} is already a member of group {}".format(user.name, group.name))
                    else:
                        raise rError

        for user in group.users:
            print(user.name)


if __name__ == "__main__":
    main()



####
# This script sets up a server. It uploads datasources and workbooks from the local filesystem.
#
# By default, all content is published to the Default project on the Default site.
####

import argparse
import glob
import logging
import tableauserverclient as TSC


def main():
    parser = argparse.ArgumentParser(description="Initialize a server with content.")
    # Common options; please keep those in sync across all samples
    parser.add_argument("--server", "-s", required=True, help="server address")
    parser.add_argument("--site", "-S", help="site name")
    parser.add_argument(
        "--token-name", "-p", required=True, help="name of the personal access token used to sign into the server"
    )
    parser.add_argument(
        "--token-value", "-v", required=True, help="value of the personal access token used to sign into the server"
    )
    parser.add_argument(
        "--logging-level",
        "-l",
        choices=["debug", "info", "error"],
        default="error",
        help="desired logging level (set to error by default)",
    )
    # Options specific to this sample
    parser.add_argument("--datasources-folder", "-df", required=True, help="folder containing datasources")
    parser.add_argument("--workbooks-folder", "-wf", required=True, help="folder containing workbooks")
    parser.add_argument("--project", required=False, default="Default", help="project to use")

    args = parser.parse_args()

    # Set logging level based on user input, or error by default
    logging_level = getattr(logging, args.logging_level.upper())
    logging.basicConfig(level=logging_level)

    ################################################################################
    # Step 1: Sign in to server.
    ################################################################################
    tableau_auth = TSC.PersonalAccessTokenAuth(args.token_name, args.token_value, site_id=args.site)
    server = TSC.Server(args.server, use_server_version=True)
    with server.auth.sign_in(tableau_auth):
        ################################################################################
        # Step 2: Create the site we need only if it doesn't exist
        ################################################################################
        print("Checking to see if we need to create the site...")

        all_sites = TSC.Pager(server.sites)
        existing_site = next((s for s in all_sites if s.content_url == args.site_id), None)

        # Create the site if it doesn't exist
        if existing_site is None:
            print("Site not found: {0} Creating it...".format(args.site_id))
            new_site = TSC.SiteItem(
                name=args.site_id,
                content_url=args.site_id.replace(" ", ""),
                admin_mode=TSC.SiteItem.AdminMode.ContentAndUsers,
            )
            server.sites.create(new_site)
        else:
            print("Site {0} exists. Moving on...".format(args.site_id))

    ################################################################################
    # Step 3: Sign-in to our target site
    ################################################################################
    print("Starting our content upload...")
    server_upload = TSC.Server(args.server)

    tableau_auth.site_id = args.site_id

    with server_upload.auth.sign_in(tableau_auth):
        ################################################################################
        # Step 4: Create the project we need only if it doesn't exist
        ################################################################################
        import time

        time.sleep(2)  # sad panda...something about eventually consistent model
        all_projects = TSC.Pager(server_upload.projects)
        project = next((p for p in all_projects if p.name.lower() == args.project.lower()), None)

        # Create our project if it doesn't exist
        if project is None:
            print("Project not found: {0} Creating it...".format(args.project))
            new_project = TSC.ProjectItem(name=args.project)
            project = server_upload.projects.create(new_project)

        ################################################################################
        # Step 5:  Set up our content
        #     Publish datasources to our site and project
        #     Publish workbooks to our site and project
        ################################################################################
        publish_datasources_to_site(server_upload, project, args.datasources_folder)
        publish_workbooks_to_site(server_upload, project, args.workbooks_folder)


def publish_datasources_to_site(server_object, project, folder):
    path = folder + "/*.tds*"

    for fname in glob.glob(path):
        new_ds = TSC.DatasourceItem(project.id)
        new_ds = server_object.datasources.publish(new_ds, fname, server_object.PublishMode.Overwrite)
        print("Datasource published. ID: {0}".format(new_ds.id))


def publish_workbooks_to_site(server_object, project, folder):
    path = folder + "/*.twb*"

    for fname in glob.glob(path):
        new_workbook = TSC.WorkbookItem(project.id)
        new_workbook.show_tabs = True
        new_workbook = server_object.workbooks.publish(new_workbook, fname, server_object.PublishMode.Overwrite)
        print("Workbook published. ID: {0}".format(new_workbook.id))


if __name__ == "__main__":
    main()



import logging

from .endpoint import Endpoint
from .exceptions import MissingRequiredFieldError
from tableauserverclient.server import RequestFactory
from tableauserverclient.models import DQWItem

logger = logging.getLogger(__name__)


class _DataQualityWarningEndpoint(Endpoint):
    def __init__(self, parent_srv, resource_type):
        super(_DataQualityWarningEndpoint, self).__init__(parent_srv)
        self.resource_type = resource_type

    @property
    def baseurl(self):
        return "{0}/sites/{1}/dataQualityWarnings/{2}".format(
            self.parent_srv.baseurl, self.parent_srv.site_id, self.resource_type
        )

    def add(self, resource, warning):
        url = "{baseurl}/{content_luid}".format(baseurl=self.baseurl, content_luid=resource.id)
        add_req = RequestFactory.DQW.add_req(warning)
        response = self.post_request(url, add_req)
        warnings = DQWItem.from_response(response.content, self.parent_srv.namespace)
        logger.info("Added dqw for resource {0}".format(resource.id))

        return warnings

    def update(self, resource, warning):
        url = "{baseurl}/{content_luid}".format(baseurl=self.baseurl, content_luid=resource.id)
        add_req = RequestFactory.DQW.update_req(warning)
        response = self.put_request(url, add_req)
        warnings = DQWItem.from_response(response.content, self.parent_srv.namespace)
        logger.info("Added dqw for resource {0}".format(resource.id))

        return warnings

    def clear(self, resource):
        url = "{baseurl}/{content_luid}".format(baseurl=self.baseurl, content_luid=resource.id)
        return self.delete_request(url)

    def populate(self, item):
        if not item.id:
            error = "Server item is missing ID. Item must be retrieved from server first."
            raise MissingRequiredFieldError(error)

        def dqw_fetcher():
            return self._get_data_quality_warnings(item)

        item._set_data_quality_warnings(dqw_fetcher)
        logger.info("Populated permissions for item (ID: {0})".format(item.id))

    def _get_data_quality_warnings(self, item, req_options=None):
        url = "{baseurl}/{content_luid}".format(baseurl=self.baseurl, content_luid=item.id)
        server_response = self.get_request(url, req_options)
        dqws = DQWItem.from_response(server_response.content, self.parent_srv.namespace)

        return dqws



import os
import requests_mock
import unittest

from io import BytesIO

import tableauserverclient as TSC
from tableauserverclient.datetime_helpers import format_datetime
from ._utils import read_xml_asset, asset

TEST_ASSET_DIR = os.path.join(os.path.dirname(__file__), "assets")

GET_XML = os.path.join(TEST_ASSET_DIR, "flow_get.xml")
POPULATE_CONNECTIONS_XML = os.path.join(TEST_ASSET_DIR, "flow_populate_connections.xml")
POPULATE_PERMISSIONS_XML = os.path.join(TEST_ASSET_DIR, "flow_populate_permissions.xml")
PUBLISH_XML = os.path.join(TEST_ASSET_DIR, "flow_publish.xml")
UPDATE_XML = os.path.join(TEST_ASSET_DIR, "flow_update.xml")
REFRESH_XML = os.path.join(TEST_ASSET_DIR, "flow_refresh.xml")


class FlowTests(unittest.TestCase):
    def setUp(self) -> None:
        self.server = TSC.Server("http://test", False)

        # Fake signin
        self.server._site_id = "dad65087-b08b-4603-af4e-2887b8aafc67"
        self.server._auth_token = "j80k54ll2lfMZ0tv97mlPvvSCRyD0DOM"
        self.server.version = "3.5"

        self.baseurl = self.server.flows.baseurl

    def test_download(self) -> None:
        with requests_mock.mock() as m:
            m.get(
                self.baseurl + "/587daa37-b84d-4400-a9a2-aa90e0be7837/content",
                headers={"Content-Disposition": 'name="tableau_flow"; filename="FlowOne.tfl"'},
            )
            file_path = self.server.flows.download("587daa37-b84d-4400-a9a2-aa90e0be7837")
            self.assertTrue(os.path.exists(file_path))
        os.remove(file_path)

    def test_download_object(self) -> None:
        with BytesIO() as file_object:
            with requests_mock.mock() as m:
                m.get(
                    self.baseurl + "/587daa37-b84d-4400-a9a2-aa90e0be7837/content",
                    headers={"Content-Disposition": 'name="tableau_flow"; filename="FlowOne.tfl"'},
                )
                file_path = self.server.flows.download("587daa37-b84d-4400-a9a2-aa90e0be7837", filepath=file_object)
                self.assertTrue(isinstance(file_path, BytesIO))

    def test_get(self) -> None:
        response_xml = read_xml_asset(GET_XML)
        with requests_mock.mock() as m:
            m.get(self.baseurl, text=response_xml)
            all_flows, pagination_item = self.server.flows.get()

        self.assertEqual(5, pagination_item.total_available)
        self.assertEqual("587daa37-b84d-4400-a9a2-aa90e0be7837", all_flows[0].id)
        self.assertEqual("http://tableauserver/#/flows/1", all_flows[0].webpage_url)
        self.assertEqual("2019-06-16T21:43:28Z", format_datetime(all_flows[0].created_at))
        self.assertEqual("2019-06-16T21:43:28Z", format_datetime(all_flows[0].updated_at))
        self.assertEqual("Default", all_flows[0].project_name)
        self.assertEqual("FlowOne", all_flows[0].name)
        self.assertEqual("aa23f4ac-906f-11e9-86fb-3f0f71412e77", all_flows[0].project_id)
        self.assertEqual("7ebb3f20-0fd2-4f27-a2f6-c539470999e2", all_flows[0].owner_id)
        self.assertEqual({"i_love_tags"}, all_flows[0].tags)
        self.assertEqual("Descriptive", all_flows[0].description)

        self.assertEqual("5c36be69-eb30-461b-b66e-3e2a8e27cc35", all_flows[1].id)
        self.assertEqual("http://tableauserver/#/flows/4", all_flows[1].webpage_url)
        self.assertEqual("2019-06-18T03:08:19Z", format_datetime(all_flows[1].created_at))
        self.assertEqual("2019-06-18T03:08:19Z", format_datetime(all_flows[1].updated_at))
        self.assertEqual("Default", all_flows[1].project_name)
        self.assertEqual("FlowTwo", all_flows[1].name)
        self.assertEqual("aa23f4ac-906f-11e9-86fb-3f0f71412e77", all_flows[1].project_id)
        self.assertEqual("9127d03f-d996-405f-b392-631b25183a0f", all_flows[1].owner_id)

    def test_update(self) -> None:
        response_xml = read_xml_asset(UPDATE_XML)
        with requests_mock.mock() as m:
            m.put(self.baseurl + "/587daa37-b84d-4400-a9a2-aa90e0be7837", text=response_xml)
            single_datasource = TSC.FlowItem("test", "aa23f4ac-906f-11e9-86fb-3f0f71412e77")
            single_datasource.owner_id = "7ebb3f20-0fd2-4f27-a2f6-c539470999e2"
            single_datasource._id = "587daa37-b84d-4400-a9a2-aa90e0be7837"
            single_datasource.description = "So fun to see"
            single_datasource = self.server.flows.update(single_datasource)

        self.assertEqual("587daa37-b84d-4400-a9a2-aa90e0be7837", single_datasource.id)
        self.assertEqual("aa23f4ac-906f-11e9-86fb-3f0f71412e77", single_datasource.project_id)
        self.assertEqual("7ebb3f20-0fd2-4f27-a2f6-c539470999e2", single_datasource.owner_id)
        self.assertEqual("So fun to see", single_datasource.description)

    def test_populate_connections(self) -> None:
        response_xml = read_xml_asset(POPULATE_CONNECTIONS_XML)
        with requests_mock.mock() as m:
            m.get(self.baseurl + "/9dbd2263-16b5-46e1-9c43-a76bb8ab65fb/connections", text=response_xml)
            single_datasource = TSC.FlowItem("test", "aa23f4ac-906f-11e9-86fb-3f0f71412e77")
            single_datasource.owner_id = "dd2239f6-ddf1-4107-981a-4cf94e415794"
            single_datasource._id = "9dbd2263-16b5-46e1-9c43-a76bb8ab65fb"
            self.server.flows.populate_connections(single_datasource)
            self.assertEqual("9dbd2263-16b5-46e1-9c43-a76bb8ab65fb", single_datasource.id)
            connections = single_datasource.connections

        self.assertTrue(connections)
        conn1, conn2, conn3 = connections
        self.assertEqual("405c1e4b-60c9-499f-9c47-a4ef1af69359", conn1.id)
        self.assertEqual("excel-direct", conn1.connection_type)
        self.assertEqual("", conn1.server_address)
        self.assertEqual("", conn1.username)
        self.assertEqual(False, conn1.embed_password)
        self.assertEqual("b47f41b1-2c47-41a3-8b17-a38ebe8b340c", conn2.id)
        self.assertEqual("sqlserver", conn2.connection_type)
        self.assertEqual("test.database.com", conn2.server_address)
        self.assertEqual("bob", conn2.username)
        self.assertEqual(False, conn2.embed_password)
        self.assertEqual("4f4a3b78-0554-43a7-b327-9605e9df9dd2", conn3.id)
        self.assertEqual("tableau-server-site", conn3.connection_type)
        self.assertEqual("http://tableauserver", conn3.server_address)
        self.assertEqual("sally", conn3.username)
        self.assertEqual(True, conn3.embed_password)

    def test_populate_permissions(self) -> None:
        with open(asset(POPULATE_PERMISSIONS_XML), "rb") as f:
            response_xml = f.read().decode("utf-8")
        with requests_mock.mock() as m:
            m.get(self.baseurl + "/0448d2ed-590d-4fa0-b272-a2a8a24555b5/permissions", text=response_xml)
            single_datasource = TSC.FlowItem("test")
            single_datasource._id = "0448d2ed-590d-4fa0-b272-a2a8a24555b5"

            self.server.flows.populate_permissions(single_datasource)
            permissions = single_datasource.permissions

            self.assertEqual(permissions[0].grantee.tag_name, "group")
            self.assertEqual(permissions[0].grantee.id, "aa42f384-906f-11e9-86fc-bb24278874b9")
            self.assertDictEqual(
                permissions[0].capabilities,
                {
                    TSC.Permission.Capability.Write: TSC.Permission.Mode.Allow,
                    TSC.Permission.Capability.Read: TSC.Permission.Mode.Allow,
                },
            )

    def test_publish(self) -> None:
        with open(PUBLISH_XML, "rb") as f:
            response_xml = f.read().decode("utf-8")
        with requests_mock.mock() as m:
            m.post(self.baseurl, text=response_xml)

            new_flow = TSC.FlowItem(name="SampleFlow", project_id="ee8c6e70-43b6-11e6-af4f-f7b0d8e20760")

            sample_flow = os.path.join(TEST_ASSET_DIR, "SampleFlow.tfl")
            publish_mode = self.server.PublishMode.CreateNew

            new_flow = self.server.flows.publish(new_flow, sample_flow, publish_mode)

        self.assertEqual("2457c468-1b24-461a-8f95-a461b3209d32", new_flow.id)
        self.assertEqual("SampleFlow", new_flow.name)
        self.assertEqual("2023-01-13T09:50:55Z", format_datetime(new_flow.created_at))
        self.assertEqual("2023-01-13T09:50:55Z", format_datetime(new_flow.updated_at))
        self.assertEqual("ee8c6e70-43b6-11e6-af4f-f7b0d8e20760", new_flow.project_id)
        self.assertEqual("default", new_flow.project_name)
        self.assertEqual("5de011f8-5aa9-4d5b-b991-f462c8dd6bb7", new_flow.owner_id)

    def test_publish_file_object(self) -> None:
        with open(PUBLISH_XML, "rb") as f:
            response_xml = f.read().decode("utf-8")
        with requests_mock.mock() as m:
            m.post(self.baseurl, text=response_xml)

            new_flow = TSC.FlowItem(name="SampleFlow", project_id="ee8c6e70-43b6-11e6-af4f-f7b0d8e20760")

            sample_flow = os.path.join(TEST_ASSET_DIR, "SampleFlow.tfl")
            publish_mode = self.server.PublishMode.CreateNew

            with open(sample_flow, "rb") as fp:
                publish_mode = self.server.PublishMode.CreateNew

                new_flow = self.server.flows.publish(new_flow, fp, publish_mode)

        self.assertEqual("2457c468-1b24-461a-8f95-a461b3209d32", new_flow.id)
        self.assertEqual("SampleFlow", new_flow.name)
        self.assertEqual("2023-01-13T09:50:55Z", format_datetime(new_flow.created_at))
        self.assertEqual("2023-01-13T09:50:55Z", format_datetime(new_flow.updated_at))
        self.assertEqual("ee8c6e70-43b6-11e6-af4f-f7b0d8e20760", new_flow.project_id)
        self.assertEqual("default", new_flow.project_name)
        self.assertEqual("5de011f8-5aa9-4d5b-b991-f462c8dd6bb7", new_flow.owner_id)

    def test_refresh(self):
        with open(asset(REFRESH_XML), "rb") as f:
            response_xml = f.read().decode("utf-8")
        with requests_mock.mock() as m:
            m.post(self.baseurl + "/92967d2d-c7e2-46d0-8847-4802df58f484/run", text=response_xml)
            flow_item = TSC.FlowItem("test")
            flow_item._id = "92967d2d-c7e2-46d0-8847-4802df58f484"
            refresh_job = self.server.flows.refresh(flow_item)

            self.assertEqual(refresh_job.id, "d1b2ccd0-6dfa-444a-aee4-723dbd6b7c9d")
            self.assertEqual(refresh_job.mode, "Asynchronous")
            self.assertEqual(refresh_job.type, "RunFlow")
            self.assertEqual(format_datetime(refresh_job.created_at), "2018-05-22T13:00:29Z")
            self.assertIsInstance(refresh_job.flow_run, TSC.FlowRunItem)
            self.assertEqual(refresh_job.flow_run.id, "e0c3067f-2333-4eee-8028-e0a56ca496f6")
            self.assertEqual(refresh_job.flow_run.flow_id, "92967d2d-c7e2-46d0-8847-4802df58f484")
            self.assertEqual(format_datetime(refresh_job.flow_run.started_at), "2018-05-22T13:00:29Z")



####
# This script demonstrates how to use the Tableau Server Client
# to filter and sort on the name of the projects present on site.
#
# To run the script, you must have installed Python 3.7 or later.
####

import argparse
import logging
import urllib.parse

import tableauserverclient as TSC


def create_example_project(
    name="Example Project",
    content_permissions="LockedToProject",
    description="Project created for testing",
    server=None,
):
    new_project = TSC.ProjectItem(name=name, content_permissions=content_permissions, description=description)
    try:
        server.projects.create(new_project)
        print("Created a new project called: %s" % name)
    except TSC.ServerResponseError:
        print("We have already created this resource: %s" % name)


def main():
    parser = argparse.ArgumentParser(description="Filter and sort projects.")
    # Common options; please keep those in sync across all samples
    parser.add_argument("--server", "-s", required=True, help="server address")
    parser.add_argument("--site", "-S", help="site name")
    parser.add_argument(
        "--token-name", "-p", required=True, help="name of the personal access token used to sign into the server"
    )
    parser.add_argument(
        "--token-value", "-v", required=True, help="value of the personal access token used to sign into the server"
    )
    parser.add_argument(
        "--logging-level",
        "-l",
        choices=["debug", "info", "error"],
        default="error",
        help="desired logging level (set to error by default)",
    )
    # Options specific to this sample
    # This sample has no additional options, yet. If you add some, please add them here

    args = parser.parse_args()

    # Set logging level based on user input, or error by default
    logging_level = getattr(logging, args.logging_level.upper())
    logging.basicConfig(level=logging_level)

    tableau_auth = TSC.PersonalAccessTokenAuth(args.token_name, args.token_value, site_id=args.site)
    server = TSC.Server(args.server, use_server_version=True)
    with server.auth.sign_in(tableau_auth):
        # Use highest Server REST API version available
        server.use_server_version()

        filter_project_name = "default"
        options = TSC.RequestOptions()

        options.filter.add(
            TSC.Filter(TSC.RequestOptions.Field.Name, TSC.RequestOptions.Operator.Equals, filter_project_name)
        )

        filtered_projects, _ = server.projects.get(req_options=options)
        # Result can either be a matching project or an empty list
        if filtered_projects:
            project_name = filtered_projects.pop().name
            print(project_name)
        else:
            error = "No project named '{}' found".format(filter_project_name)
            print(error)

        create_example_project(name="Example 1", server=server)
        create_example_project(name="Example 2", server=server)
        create_example_project(name="Example 3", server=server)
        create_example_project(name="Proiect ca Exemplu", server=server)

        options = TSC.RequestOptions()

        # don't forget to URL encode the query names
        options.filter.add(
            TSC.Filter(
                TSC.RequestOptions.Field.Name, TSC.RequestOptions.Operator.In, ["Example+1", "Example+2", "Example+3"]
            )
        )

        options.sort.add(TSC.Sort(TSC.RequestOptions.Field.Name, TSC.RequestOptions.Direction.Desc))

        matching_projects, pagination_item = server.projects.get(req_options=options)
        print("Filtered projects are:")
        for project in matching_projects:
            print(project.name, project.id)

        # Or, try the django style filtering.
        projects = ["Example 1", "Example 2", "Example 3"]
        projects = [urllib.parse.quote_plus(p) for p in projects]
        for project in server.projects.filter(name__in=projects).sort("-name"):
            print(project.name, project.id)


if __name__ == "__main__":
    main()



import logging

from .dqw_endpoint import _DataQualityWarningEndpoint
from .endpoint import api, Endpoint
from .exceptions import MissingRequiredFieldError
from .permissions_endpoint import _PermissionsEndpoint
from tableauserverclient.server import RequestFactory
from tableauserverclient.models import TableItem, ColumnItem, PaginationItem
from ..pager import Pager

logger = logging.getLogger("tableau.endpoint.tables")


class Tables(Endpoint):
    def __init__(self, parent_srv):
        super(Tables, self).__init__(parent_srv)

        self._permissions = _PermissionsEndpoint(parent_srv, lambda: self.baseurl)
        self._data_quality_warnings = _DataQualityWarningEndpoint(self.parent_srv, "table")

    @property
    def baseurl(self):
        return "{0}/sites/{1}/tables".format(self.parent_srv.baseurl, self.parent_srv.site_id)

    @api(version="3.5")
    def get(self, req_options=None):
        logger.info("Querying all tables on site")
        url = self.baseurl
        server_response = self.get_request(url, req_options)
        pagination_item = PaginationItem.from_response(server_response.content, self.parent_srv.namespace)
        all_table_items = TableItem.from_response(server_response.content, self.parent_srv.namespace)
        return all_table_items, pagination_item

    # Get 1 table
    @api(version="3.5")
    def get_by_id(self, table_id):
        if not table_id:
            error = "table ID undefined."
            raise ValueError(error)
        logger.info("Querying single table (ID: {0})".format(table_id))
        url = "{0}/{1}".format(self.baseurl, table_id)
        server_response = self.get_request(url)
        return TableItem.from_response(server_response.content, self.parent_srv.namespace)[0]

    @api(version="3.5")
    def delete(self, table_id):
        if not table_id:
            error = "Database ID undefined."
            raise ValueError(error)
        url = "{0}/{1}".format(self.baseurl, table_id)
        self.delete_request(url)
        logger.info("Deleted single table (ID: {0})".format(table_id))

    @api(version="3.5")
    def update(self, table_item):
        if not table_item.id:
            error = "table item missing ID."
            raise MissingRequiredFieldError(error)

        url = "{0}/{1}".format(self.baseurl, table_item.id)
        update_req = RequestFactory.Table.update_req(table_item)
        server_response = self.put_request(url, update_req)
        logger.info("Updated table item (ID: {0})".format(table_item.id))
        updated_table = TableItem.from_response(server_response.content, self.parent_srv.namespace)[0]
        return updated_table

    # Get all columns of the table
    @api(version="3.5")
    def populate_columns(self, table_item, req_options=None):
        if not table_item.id:
            error = "Table item missing ID. table must be retrieved from server first."
            raise MissingRequiredFieldError(error)

        def column_fetcher():
            return Pager(
                lambda options: self._get_columns_for_table(table_item, options),
                req_options,
            )

        table_item._set_columns(column_fetcher)
        logger.info("Populated columns for table (ID: {0}".format(table_item.id))

    def _get_columns_for_table(self, table_item, req_options=None):
        url = "{0}/{1}/columns".format(self.baseurl, table_item.id)
        server_response = self.get_request(url, req_options)
        columns = ColumnItem.from_response(server_response.content, self.parent_srv.namespace)
        pagination_item = PaginationItem.from_response(server_response.content, self.parent_srv.namespace)
        return columns, pagination_item

    @api(version="3.5")
    def update_column(self, table_item, column_item):
        url = "{0}/{1}/columns/{2}".format(self.baseurl, table_item.id, column_item.id)
        update_req = RequestFactory.Column.update_req(column_item)
        server_response = self.put_request(url, update_req)
        column = ColumnItem.from_response(server_response.content, self.parent_srv.namespace)[0]

        logger.info("Updated table item (ID: {0} & column item {1}".format(table_item.id, column_item.id))
        return column

    @api(version="3.5")
    def populate_permissions(self, item):
        self._permissions.populate(item)

    @api(version="3.5")
    def update_permission(self, item, rules):
        import warnings

        warnings.warn(
            "Server.tables.update_permission is deprecated, " "please use Server.tables.update_permissions instead.",
            DeprecationWarning,
        )
        return self._permissions.update(item, rules)

    @api(version="3.5")
    def update_permissions(self, item, rules):
        return self._permissions.update(item, rules)

    @api(version="3.5")
    def delete_permission(self, item, rules):
        return self._permissions.delete(item, rules)

    @api(version="3.5")
    def populate_dqw(self, item):
        self._data_quality_warnings.populate(item)

    @api(version="3.5")
    def update_dqw(self, item, warning):
        return self._data_quality_warnings.update(item, warning)

    @api(version="3.5")
    def add_dqw(self, item, warning):
        return self._data_quality_warnings.add(item, warning)

    @api(version="3.5")
    def delete_dqw(self, item):
        self._data_quality_warnings.clear(item)



import datetime
import re
from functools import wraps

from tableauserverclient.datetime_helpers import parse_datetime


def property_is_enum(enum_type):
    def property_type_decorator(func):
        @wraps(func)
        def wrapper(self, value):
            if value is not None and not hasattr(enum_type, value):
                error = "Invalid value: {0}. {1} must be of type {2}.".format(value, func.__name__, enum_type.__name__)
                raise ValueError(error)
            return func(self, value)

        return wrapper

    return property_type_decorator


def property_is_boolean(func):
    @wraps(func)
    def wrapper(self, value):
        if not isinstance(value, bool):
            error = "Boolean expected for {0} flag.".format(func.__name__)
            raise ValueError(error)
        return func(self, value)

    return wrapper


def property_not_nullable(func):
    @wraps(func)
    def wrapper(self, value):
        if value is None:
            error = "{0} must be defined.".format(func.__name__)
            raise ValueError(error)
        return func(self, value)

    return wrapper


def property_not_empty(func):
    @wraps(func)
    def wrapper(self, value):
        if not value:
            error = "{0} must not be empty.".format(func.__name__)
            raise ValueError(error)
        return func(self, value)

    return wrapper


def property_is_valid_time(func):
    @wraps(func)
    def wrapper(self, value):
        units_of_time = {"hour", "minute", "second"}

        if not any(hasattr(value, unit) for unit in units_of_time):
            error = "Invalid time object defined."
            raise ValueError(error)
        return func(self, value)

    return wrapper


def property_is_int(range, allowed=None):
    """Takes a range of ints and a list of exemptions to check against
    when setting a property on a model. The range is a tuple of (min, max) and the
    allowed list (empty by default) allows values outside that range.
    This is useful for when we use sentinel values.

    Example: Revisions allow a range of 2-10000, but use -1 as a sentinel for 'unlimited'.
    """

    if allowed is None:
        allowed = ()  # Empty tuple for fast no-op testing.

    def property_type_decorator(func):
        @wraps(func)
        def wrapper(self, value):
            error = "Invalid property defined: '{}'. Integer value expected.".format(value)

            if range is None:
                if isinstance(value, int):
                    return func(self, value)
                else:
                    raise ValueError(error)

            min, max = range

            if (value < min or value > max) and (value not in allowed):
                raise ValueError(error)

            return func(self, value)

        return wrapper

    return property_type_decorator


def property_matches(regex_to_match, error):
    compiled_re = re.compile(regex_to_match)

    def wrapper(func):
        @wraps(func)
        def validate_regex_decorator(self, value):
            if not compiled_re.match(value):
                raise ValueError(error)
            return func(self, value)

        return validate_regex_decorator

    return wrapper


def property_is_datetime(func):
    """Takes the following datetime format and turns it into a datetime object:

    2016-08-18T18:25:36Z

    Because we return everything with Z as the timezone, we assume everything is in UTC and create
    a timezone aware datetime.
    """

    @wraps(func)
    def wrapper(self, value):
        if isinstance(value, datetime.datetime):
            return func(self, value)
        if not isinstance(value, str):
            raise ValueError(
                "Cannot convert {} into a datetime, cannot update {}".format(value.__class__.__name__, func.__name__)
            )

        dt = parse_datetime(value)
        return func(self, dt)

    return wrapper


def property_is_data_acceleration_config(func):
    @wraps(func)
    def wrapper(self, value):
        if not isinstance(value, dict):
            raise ValueError("{} is not type 'dict', cannot update {})".format(value.__class__.__name__, func.__name__))
        if len(value) != 4 or not all(
            attr in value.keys()
            for attr in (
                "acceleration_enabled",
                "accelerate_now",
                "last_updated_at",
                "acceleration_status",
            )
        ):
            error = "{} should have 2 keys ".format(func.__name__)
            error += "'acceleration_enabled' and 'accelerate_now'"
            error += "instead you have {}".format(value.keys())
            raise ValueError(error)
        return func(self, value)

    return wrapper



from .property_decorators import property_is_boolean


class ConnectionCredentials(object):
    """Connection Credentials for Workbooks and Datasources publish request.

    Consider removing this object and other variables holding secrets
    as soon as possible after use to avoid them hanging around in memory.

    """

    def __init__(self, name, password, embed=True, oauth=False):
        self.name = name
        self.password = password
        self.embed = embed
        self.oauth = oauth

    @property
    def embed(self):
        return self._embed

    @embed.setter
    @property_is_boolean
    def embed(self, value):
        self._embed = value

    @property
    def oauth(self):
        return self._oauth

    @oauth.setter
    @property_is_boolean
    def oauth(self, value):
        self._oauth = value

    @classmethod
    def from_xml_element(cls, parsed_response, ns):
        connection_creds_xml = parsed_response.find(".//t:connectionCredentials", namespaces=ns)

        name = connection_creds_xml.get("name", None)
        password = connection_creds_xml.get("password", None)
        embed = connection_creds_xml.get("embed", None)
        oAuth = connection_creds_xml.get("oAuth", None)

        connection_creds = cls(name, password, embed, oAuth)
        return connection_creds



import pytest

from unittest import TestCase
import _models


# ensure that all models have a __repr__ method implemented
class TestAllModels(TestCase):

    """
    ColumnItem wrapper_descriptor
    ConnectionCredentials wrapper_descriptor
    DataAccelerationReportItem wrapper_descriptor
    DatabaseItem wrapper_descriptor
    DQWItem wrapper_descriptor
    UnpopulatedPropertyError wrapper_descriptor
    FavoriteItem wrapper_descriptor
    FlowRunItem wrapper_descriptor
    IntervalItem wrapper_descriptor
    DailyInterval wrapper_descriptor
    WeeklyInterval wrapper_descriptor
    MonthlyInterval wrapper_descriptor
    HourlyInterval wrapper_descriptor
    BackgroundJobItem wrapper_descriptor
    PaginationItem wrapper_descriptor
    Permission wrapper_descriptor
    ServerInfoItem wrapper_descriptor
    SiteItem wrapper_descriptor
    TableItem wrapper_descriptor
    Resource wrapper_descriptor
    """

    # not all models have __repr__ yet: see above list
    @pytest.mark.xfail()
    def test_repr_is_implemented(self):
        m = _models.get_defined_models()
        for model in m:
            with self.subTest(model.__name__, model=model):
                print(model.__name__, type(model.__repr__).__name__)
                self.assertEqual(type(model.__repr__).__name__, "function")



# encoding=utf-8
import unittest
import os
import requests_mock
import tableauserverclient as TSC
from tableauserverclient.datetime_helpers import format_datetime

TEST_ASSET_DIR = os.path.join(os.path.dirname(__file__), "assets")

GET_XML = os.path.join(TEST_ASSET_DIR, "group_get.xml")
POPULATE_USERS = os.path.join(TEST_ASSET_DIR, "group_populate_users.xml")
POPULATE_USERS_EMPTY = os.path.join(TEST_ASSET_DIR, "group_populate_users_empty.xml")
ADD_USER = os.path.join(TEST_ASSET_DIR, "group_add_user.xml")
ADD_USER_POPULATE = os.path.join(TEST_ASSET_DIR, "group_users_added.xml")
CREATE_GROUP = os.path.join(TEST_ASSET_DIR, "group_create.xml")
CREATE_GROUP_AD = os.path.join(TEST_ASSET_DIR, "group_create_ad.xml")
CREATE_GROUP_ASYNC = os.path.join(TEST_ASSET_DIR, "group_create_async.xml")
UPDATE_XML = os.path.join(TEST_ASSET_DIR, "group_update.xml")


class GroupTests(unittest.TestCase):
    def setUp(self) -> None:
        self.server = TSC.Server("http://test", False)

        # Fake signin
        self.server._site_id = "dad65087-b08b-4603-af4e-2887b8aafc67"
        self.server._auth_token = "j80k54ll2lfMZ0tv97mlPvvSCRyD0DOM"

        self.baseurl = self.server.groups.baseurl

    def test_get(self) -> None:
        with open(GET_XML, "rb") as f:
            response_xml = f.read().decode("utf-8")
        with requests_mock.mock() as m:
            m.get(self.baseurl, text=response_xml)
            all_groups, pagination_item = self.server.groups.get()

        self.assertEqual(3, pagination_item.total_available)
        self.assertEqual("ef8b19c0-43b6-11e6-af50-63f5805dbe3c", all_groups[0].id)
        self.assertEqual("All Users", all_groups[0].name)
        self.assertEqual("local", all_groups[0].domain_name)

        self.assertEqual("e7833b48-c6f7-47b5-a2a7-36e7dd232758", all_groups[1].id)
        self.assertEqual("Another group", all_groups[1].name)
        self.assertEqual("local", all_groups[1].domain_name)

        self.assertEqual("86a66d40-f289-472a-83d0-927b0f954dc8", all_groups[2].id)
        self.assertEqual("TableauExample", all_groups[2].name)
        self.assertEqual("local", all_groups[2].domain_name)

    def test_get_before_signin(self) -> None:
        self.server._auth_token = None
        self.assertRaises(TSC.NotSignedInError, self.server.groups.get)

    def test_populate_users(self) -> None:
        with open(POPULATE_USERS, "rb") as f:
            response_xml = f.read().decode("utf-8")
        with requests_mock.mock() as m:
            m.get(
                self.baseurl + "/e7833b48-c6f7-47b5-a2a7-36e7dd232758/users?pageNumber=1&pageSize=100",
                text=response_xml,
                complete_qs=True,
            )
            single_group = TSC.GroupItem(name="Test Group")
            single_group._id = "e7833b48-c6f7-47b5-a2a7-36e7dd232758"
            self.server.groups.populate_users(single_group)

            self.assertEqual(1, len(list(single_group.users)))
            user = list(single_group.users).pop()
            self.assertEqual("dd2239f6-ddf1-4107-981a-4cf94e415794", user.id)
            self.assertEqual("alice", user.name)
            self.assertEqual("Publisher", user.site_role)
            self.assertEqual("2016-08-16T23:17:06Z", format_datetime(user.last_login))

    def test_delete(self) -> None:
        with requests_mock.mock() as m:
            m.delete(self.baseurl + "/e7833b48-c6f7-47b5-a2a7-36e7dd232758", status_code=204)
            self.server.groups.delete("e7833b48-c6f7-47b5-a2a7-36e7dd232758")

    def test_remove_user(self) -> None:
        with open(POPULATE_USERS, "rb") as f:
            response_xml_populate = f.read().decode("utf-8")

        with open(POPULATE_USERS_EMPTY, "rb") as f:
            response_xml_empty = f.read().decode("utf-8")

        with requests_mock.mock() as m:
            url = self.baseurl + "/e7833b48-c6f7-47b5-a2a7-36e7dd232758/users" "/dd2239f6-ddf1-4107-981a-4cf94e415794"

            m.delete(url, status_code=204)
            #  We register the get endpoint twice. The first time we have 1 user, the second we have 'removed' them.
            m.get(self.baseurl + "/e7833b48-c6f7-47b5-a2a7-36e7dd232758/users", text=response_xml_populate)

            single_group = TSC.GroupItem("test")
            single_group._id = "e7833b48-c6f7-47b5-a2a7-36e7dd232758"
            self.server.groups.populate_users(single_group)
            self.assertEqual(1, len(list(single_group.users)))
            self.server.groups.remove_user(single_group, "dd2239f6-ddf1-4107-981a-4cf94e415794")

            m.get(self.baseurl + "/e7833b48-c6f7-47b5-a2a7-36e7dd232758/users", text=response_xml_empty)
            self.assertEqual(0, len(list(single_group.users)))

    def test_add_user(self) -> None:
        with open(ADD_USER, "rb") as f:
            response_xml_add = f.read().decode("utf-8")
        with open(ADD_USER_POPULATE, "rb") as f:
            response_xml_populate = f.read().decode("utf-8")
        with requests_mock.mock() as m:
            m.post(self.baseurl + "/e7833b48-c6f7-47b5-a2a7-36e7dd232758/users", text=response_xml_add)
            m.get(self.baseurl + "/e7833b48-c6f7-47b5-a2a7-36e7dd232758/users", text=response_xml_populate)
            single_group = TSC.GroupItem("test")
            single_group._id = "e7833b48-c6f7-47b5-a2a7-36e7dd232758"

            self.server.groups.add_user(single_group, "5de011f8-5aa9-4d5b-b991-f462c8dd6bb7")
            self.server.groups.populate_users(single_group)
            self.assertEqual(1, len(list(single_group.users)))
            user = list(single_group.users).pop()
            self.assertEqual("5de011f8-5aa9-4d5b-b991-f462c8dd6bb7", user.id)
            self.assertEqual("testuser", user.name)
            self.assertEqual("ServerAdministrator", user.site_role)

    def test_add_user_before_populating(self) -> None:
        with open(GET_XML, "rb") as f:
            get_xml_response = f.read().decode("utf-8")
        with open(ADD_USER, "rb") as f:
            add_user_response = f.read().decode("utf-8")
        with requests_mock.mock() as m:
            m.get(self.baseurl, text=get_xml_response)
            m.post(
                self.baseurl + "/ef8b19c0-43b6-11e6-af50-63f5805dbe3c/users",
                text=add_user_response,
            )
            all_groups, pagination_item = self.server.groups.get()
            single_group = all_groups[0]
            self.server.groups.add_user(single_group, "5de011f8-5aa9-4d5b-b991-f462c8dd6bb7")

    def test_add_user_missing_user_id(self) -> None:
        with open(POPULATE_USERS, "rb") as f:
            response_xml = f.read().decode("utf-8")
        with requests_mock.mock() as m:
            m.get(self.baseurl + "/e7833b48-c6f7-47b5-a2a7-36e7dd232758/users", text=response_xml)
            single_group = TSC.GroupItem(name="Test Group")
            single_group._id = "e7833b48-c6f7-47b5-a2a7-36e7dd232758"
            self.server.groups.populate_users(single_group)

        self.assertRaises(ValueError, self.server.groups.add_user, single_group, "")

    def test_add_user_missing_group_id(self) -> None:
        single_group = TSC.GroupItem("test")
        self.assertRaises(
            TSC.MissingRequiredFieldError,
            self.server.groups.add_user,
            single_group,
            "5de011f8-5aa9-4d5b-b991-f462c8dd6bb7",
        )

    def test_remove_user_before_populating(self) -> None:
        with open(GET_XML, "rb") as f:
            response_xml = f.read().decode("utf-8")
        with requests_mock.mock() as m:
            m.get(self.baseurl, text=response_xml)
            m.delete(
                self.baseurl + "/ef8b19c0-43b6-11e6-af50-63f5805dbe3c/users/5de011f8-5aa9-4d5b-b991-f462c8dd6bb7",
                text="ok",
            )
            all_groups, pagination_item = self.server.groups.get()
            single_group = all_groups[0]
            self.server.groups.remove_user(single_group, "5de011f8-5aa9-4d5b-b991-f462c8dd6bb7")

    def test_remove_user_missing_user_id(self) -> None:
        with open(POPULATE_USERS, "rb") as f:
            response_xml = f.read().decode("utf-8")
        with requests_mock.mock() as m:
            m.get(self.baseurl + "/e7833b48-c6f7-47b5-a2a7-36e7dd232758/users", text=response_xml)
            single_group = TSC.GroupItem(name="Test Group")
            single_group._id = "e7833b48-c6f7-47b5-a2a7-36e7dd232758"
            self.server.groups.populate_users(single_group)

        self.assertRaises(ValueError, self.server.groups.remove_user, single_group, "")

    def test_remove_user_missing_group_id(self) -> None:
        single_group = TSC.GroupItem("test")
        self.assertRaises(
            TSC.MissingRequiredFieldError,
            self.server.groups.remove_user,
            single_group,
            "5de011f8-5aa9-4d5b-b991-f462c8dd6bb7",
        )

    def test_create_group(self) -> None:
        with open(CREATE_GROUP, "rb") as f:
            response_xml = f.read().decode("utf-8")
        with requests_mock.mock() as m:
            m.post(self.baseurl, text=response_xml)
            group_to_create = TSC.GroupItem("試供品")
            group = self.server.groups.create(group_to_create)
            self.assertEqual(group.name, "試供品")
            self.assertEqual(group.id, "3e4a9ea0-a07a-4fe6-b50f-c345c8c81034")

    def test_create_ad_group(self) -> None:
        with open(CREATE_GROUP_AD, "rb") as f:
            response_xml = f.read().decode("utf-8")
        with requests_mock.mock() as m:
            m.post(self.baseurl, text=response_xml)
            group_to_create = TSC.GroupItem("試供品")
            group_to_create.domain_name = "just-has-to-exist"
            group = self.server.groups.create_AD_group(group_to_create, False)
            self.assertEqual(group.name, "試供品")
            self.assertEqual(group.license_mode, "onLogin")
            self.assertEqual(group.minimum_site_role, "Creator")
            self.assertEqual(group.domain_name, "active-directory-domain-name")

    def test_create_group_async(self) -> None:
        with open(CREATE_GROUP_ASYNC, "rb") as f:
            response_xml = f.read().decode("utf-8")
        with requests_mock.mock() as m:
            m.post(self.baseurl, text=response_xml)
            group_to_create = TSC.GroupItem("試供品")
            group_to_create.domain_name = "woohoo"
            job = self.server.groups.create_AD_group(group_to_create, True)
            self.assertEqual(job.mode, "Asynchronous")
            self.assertEqual(job.type, "GroupImport")

    def test_update(self) -> None:
        with open(UPDATE_XML, "rb") as f:
            response_xml = f.read().decode("utf-8")
        with requests_mock.mock() as m:
            m.put(self.baseurl + "/ef8b19c0-43b6-11e6-af50-63f5805dbe3c", text=response_xml)
            group = TSC.GroupItem(name="Test Group")
            group._domain_name = "local"
            group._id = "ef8b19c0-43b6-11e6-af50-63f5805dbe3c"
            group = self.server.groups.update(group)

        self.assertEqual("ef8b19c0-43b6-11e6-af50-63f5805dbe3c", group.id)
        self.assertEqual("Group updated name", group.name)
        self.assertEqual("ExplorerCanPublish", group.minimum_site_role)
        self.assertEqual("onLogin", group.license_mode)

    # async update is not supported for local groups
    def test_update_local_async(self) -> None:
        group = TSC.GroupItem("myGroup")
        group._id = "ef8b19c0-43b6-11e6-af50-63f5805dbe3c"
        self.assertRaises(ValueError, self.server.groups.update, group, as_job=True)

        # mimic group returned from server where domain name is set to 'local'
        group.domain_name = "local"
        self.assertRaises(ValueError, self.server.groups.update, group, as_job=True)
